from urllib.parse import quote_plus

import requests

from ._utils import (
    _remove_slash_url,
)
from .auth import access_token
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def approve_probation(
    project: str, asset: str, version: str, url: str = rest_url(), token: str = None
):
    """Approve a probational upload.

    This removes the ``on_probation`` tag from the uploaded version.

    See Also:
        :py:func:`~gypsum_client.upload_api_operations.start_upload`,
        to specify probational upload.

        :py:func:`~.reject_probation`,
        to reject the probational upload..

    Example:

        .. code-block:: python

            init = start_upload(
                project="test-Py",
                asset="probation",
                version="v1",
                files=[],
                probation=True
            )

            complete_upload(init)
            approve_probation("test-Py", "probation", "v1")

            # Cleanup if this is just for testing
            remove_asset("test-Py", "probation")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.
    """

    if token is None:
        token = access_token()

    url = _remove_slash_url(url)
    _key = f"{quote_plus(project)}/{quote_plus(asset)}/{quote_plus(version)}"
    req = requests.post(
        f"{url}/probation/approve/{_key}",
        headers={"Authorization": f"Bearer {token}"},
    )

    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to approve probation, {req.status_code} and reason: {req.text}."
        ) from e


def reject_probation(
    project: str, asset: str, version: str, url: str = rest_url(), token: str = None
):
    """Reject a probational upload.

    This removes all files associated with that version.

    See Also:
        :py:func:`~gypsum_client.upload_api_operations.start_upload`,
        to specify probational upload.

        :py:func:`~.approve_probation`,
        to approve the probational upload..

    Example:

        .. code-block:: python

            init = start_upload(
                project="test-Py",
                asset="probation",
                version="v1",
                files=[],
                probation=True
            )

            complete_upload(init)
            reject_probation("test-Py", "probation", "v1")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.
    """

    if token is None:
        token = access_token()

    url = _remove_slash_url(url)
    _key = f"{quote_plus(project)}/{quote_plus(asset)}/{quote_plus(version)}"
    req = requests.post(
        f"{url}/probation/reject/{_key}",
        headers={"Authorization": f"Bearer {token}"},
    )

    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to reject probation, {req.status_code} and reason: {req.text}."
        ) from e



import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "gypsum-client"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


from ._utils import BUCKET_CACHE_NAME
from .auth import access_token, set_access_token
from .cache_directory import cache_directory
from .clone_operations import clone_version
from .config import REQUESTS_MOD
from .create_operations import create_project
from .fetch_metadata_database import fetch_metadata_database
from .fetch_metadata_schema import fetch_metadata_schema
from .fetch_operations import (
    fetch_latest,
    fetch_manifest,
    fetch_permissions,
    fetch_quota,
    fetch_summary,
    fetch_usage,
)
from .list_operations import list_assets, list_files, list_projects, list_versions
from .prepare_directory_for_upload import prepare_directory_upload
from .probation_operations import approve_probation, reject_probation
from .refresh_operations import refresh_latest, refresh_usage
from .remove_operations import remove_asset, remove_project, remove_version
from .resolve_links import resolve_links
from .rest_url import rest_url
from .s3_config import public_s3_config
from .save_operations import save_file, save_version
from .search_metadata import define_text_query, search_metadata_text
from .set_operations import set_permissions, set_quota
from .upload_api_operations import abort_upload, complete_upload, start_upload
from .upload_file_actions import upload_directory, upload_files
from .validate_metadata import validate_metadata



import json
import os
import time
from typing import Optional

import requests
from filelock import FileLock

from .cache_directory import cache_directory
from .config import REQUESTS_MOD
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


CREDS_CACHE = {"uncached": None, "info": {}}


def _config_cache_path(cache_dir):
    return os.path.join(cache_dir, "credentials", "s3.json")


def public_s3_config(
    refresh: bool = False, url: str = rest_url(), cache_dir: Optional[str] = None
) -> dict:
    """Get S3 configuration to the bucket storing the data.

    Users can use this downstream to access the bucket directly using boto3.

    Args:
        refresh:
            Whether to refresh the cached credentials.
            Defaults to False.

        url:
            URL to the gypsum compatible API.

        cache_dir:
            Path to the cache directory.
            Defaults to None.

    Returns:
        A dictionary containing the S3 credentials.
    """
    creds = None

    if not refresh:
        if cache_dir is None:
            creds = CREDS_CACHE["uncached"]
            if creds is not None:
                return creds
        else:
            creds = CREDS_CACHE["info"].get(cache_dir)
            if creds is not None:
                return creds

            cache_path = _config_cache_path(cache_dir)
            if os.path.exists(cache_path):
                _lock = FileLock(cache_path + ".LOCK")
                with _lock:
                    with open(cache_dir, "r") as f:
                        creds = json.load(f)

                if (time.time() - os.path.getctime(cache_dir)) <= (
                    1 * 7 * 24 * 60 * 60
                ):
                    CREDS_CACHE["info"][cache_dir] = creds
                    return creds

    req = requests.get(url + "/credentials/s3-api", verify=REQUESTS_MOD["verify"])
    creds = req.json()

    if cache_dir is None:
        CREDS_CACHE["uncached"] = creds
    else:
        CREDS_CACHE["info"][cache_dir] = creds
        config_path = _config_cache_path(cache_dir)
        os.makedirs(os.path.dirname(config_path), exist_ok=True)

        with FileLock(config_path + ".LOCK"):
            with open(config_path, "w") as f:
                json.dump(creds, f)

    return creds



import os
import tempfile

import requests
from filelock import FileLock

from .cache_directory import cache_directory
from .config import REQUESTS_MOD

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def fetch_metadata_schema(
    name: str = "bioconductor/v1.json",
    cache_dir: str = cache_directory(),
    overwrite: bool = False,
) -> str:
    """Fetch a JSON schema file for metadata to be inserted into a SQLite database.

    Fetch a JSON schema file for metadata to be inserted into a SQLite database
    See `metadata index <https://github.com/ArtifactDB/bioconductor-metadata-index>`_
    for more details.

    Each SQLite database is created from metadata files uploaded to the gypsum backend,
    so clients uploading objects to be incorporated into the database should
    validate their metadata against the corresponding JSON schema.

    See Also:
        :py:func:`~gypsum_client.validate_metadata.validate_metadata`, to
        validate metadata against a chosen schema.

        :py:func:`~gypsum_client.fetch_metadata_database.fetch_metadata_database`,
        to obtain the SQLite database of metadata.

    Example:

        .. code-block:: python

            schema_path = fetch_metadata_schema()

    Args:
        name:
            Name of the schema.
            Defaults to "bioconductor/v1.json".

        cache_dir:
            Path to the cache directory.

        overwrite:
            Whether to overwrite existing file in cache.

    Returns:
        Path containing the downloaded schema.
    """
    cache_path = None
    if cache_dir is None:
        cache_path = tempfile.mktemp(suffix=".json")
    else:
        cache_dir = os.path.join(cache_dir, "schemas")

        cache_path = os.path.join(cache_dir, name)
        if not os.path.exists(cache_path):
            os.makedirs(os.path.dirname(cache_path), exist_ok=True)

        if os.path.exists(cache_path) and not overwrite:
            _lock = FileLock(cache_path + ".LOCK")
            if not _lock.is_locked:
                return cache_path

    _lock = FileLock(cache_path + ".LOCK")
    with _lock:
        url = "https://artifactdb.github.io/bioconductor-metadata-index/" + name
        response = requests.get(url, verify=REQUESTS_MOD["verify"])
        with open(cache_path, "wb") as f:
            f.write(response.content)

    return cache_path



import hashlib
import os
import shutil
import tempfile
from pathlib import Path

import pytest
from gypsum_client import (
    abort_upload,
    complete_upload,
    fetch_manifest,
    list_assets,
    remove_asset,
    start_upload,
    upload_directory,
    upload_files,
)

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

blah_contents = (
    "A\nB\nC\nD\nE\nF\nG\nH\nI\nJ\nK\nL\nM\nN\nO\nP\nQ\nR\nS\nT\nU\nV\nW\nX\nY\nZ\n"
)
foobar_contents = "1 2 3 4 5\n6 7 8 9 10\n"

app_url = "https://gypsum-test.artifactdb.com"


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_upload_regular():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_asset("test-Py", asset="upload", token=gh_token, url=app_url)
    tmp_dir = tempfile.mkdtemp()

    with open(f"{tmp_dir}/blah.txt", "w") as f:
        f.write(blah_contents)

    os.makedirs(f"{tmp_dir}/foo", exist_ok=True)

    with open(f"{tmp_dir}/foo/blah.txt", "w") as f:
        f.write(foobar_contents)

    files = [
        str(file.relative_to(tmp_dir))
        for file in Path(tmp_dir).rglob("*")
        if not os.path.isdir(file)
    ]

    init = start_upload(
        project="test-Py",
        asset="upload",
        version="1",
        files=files,
        directory=tmp_dir,
        token=gh_token,
        url=app_url,
    )

    assert len(init["file_urls"]) == 2
    assert isinstance(init["abort_url"], str)
    assert isinstance(init["complete_url"], str)
    assert isinstance(init["session_token"], str)

    upload_files(init, directory=tmp_dir, url=app_url)
    complete_upload(init, url=app_url)

    man = fetch_manifest("test-Py", "upload", "1", cache_dir=None, url=app_url)
    assert sorted(man.keys()) == ["blah.txt", "foo/blah.txt"]
    assert all(man[file].get("link") is None for file in man.keys())


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_upload_sequence_links():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_asset("test-Py", asset="upload", token=gh_token, url=app_url)

    tmp = tempfile.mkdtemp()
    try:
        fpath = os.path.join(tmp, "test.json")
        with open(fpath, "w") as f:
            f.write('[ "Aaron" ]')

        link_df = [
            {
                "from.path": "whee/stuff.txt",
                "to.project": "test-R",
                "to.asset": "basic",
                "to.version": "v1",
                "to.path": "blah.txt",
            },
            {
                "from.path": "michaela",
                "to.project": "test-R",
                "to.asset": "basic",
                "to.version": "v1",
                "to.path": "foo/bar.txt",
            },
        ]

        init = start_upload(
            project="test-Py",
            asset="upload",
            version="1",
            files=[
                {
                    "path": "test.json",
                    "size": os.path.getsize(fpath),
                    "md5sum": hashlib.md5(open(fpath, "rb").read()).hexdigest(),
                }
            ],
            links=link_df,
            directory=tmp,
            token=gh_token,
            url=app_url,
        )

        assert len(init["file_urls"]) == 1
        upload_files(init, directory=tmp, url=app_url)
        complete_upload(init, url=app_url)

        man = fetch_manifest("test-Py", "upload", "1", cache_dir=None, url=app_url)

        assert sorted(man.keys()) == ["michaela", "test.json", "whee/stuff.txt"]
        assert man["michaela"]["link"] is not None
        assert man["whee/stuff.txt"]["link"] is not None
        assert "link" not in man["test.json"]
    finally:
        shutil.rmtree(tmp)


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_aborting_upload():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_asset("test-Py", asset="upload", token=gh_token, url=app_url)

    init = start_upload(
        project="test-Py",
        asset="upload",
        version="1",
        files=[],
        token=gh_token,
        url=app_url,
    )

    assert len(init["file_urls"]) == 0
    assert "upload" in list_assets("test-Py", url=app_url)

    abort_upload(init, url=app_url)
    assert "upload" not in list_assets("test-Py", url=app_url)


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_upload_directory():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_asset("test-Py", asset="upload-dir", token=gh_token, url=app_url)

    tmp = tempfile.mkdtemp()
    try:
        with open(os.path.join(tmp, "blah.txt"), "w") as f:
            f.write("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

        os.makedirs(os.path.join(tmp, "foo"))
        with open(os.path.join(tmp, "foo", "bar.txt"), "w") as f:
            f.write("\n".join(map(str, range(1, 11))))

        upload_directory(
            tmp, "test-Py", "upload-dir", version="1", token=gh_token, url=app_url
        )

        man = fetch_manifest("test-Py", "upload-dir", "1", cache_dir=None, url=app_url)
        assert sorted(man.keys()) == ["blah.txt", "foo/bar.txt"]
        assert all("link" not in x for x in man.values())
    finally:
        shutil.rmtree(tmp)



"""Clone a version's directory structure.

Cloning of a versioned asset involves creating a directory at the destination
that has the same contents as the corresponding project-asset-version directory.
All files in the specified version are represented as symlinks from the
destination to the corresponding file in the cache.
The idea is that, when the destination is used in
:py:func:`~gypsum_client.prepare_directory_for_upload.prepare_directory_upload`,
the symlinks are converted into upload links, i.e., ``links=`` in
:py:func:`~gypsum_client.upload_api_operations.start_upload`.
This allows users to create new versions very cheaply as duplicate files
are not uploaded to/stored in the backend.

Users can more-or-less do whatever they want inside the cloned destination,
but they should treat the symlink targets as read-only.
That is, they should not modify the contents of the linked-to file, as these
refer to assumed-immutable files in the cache.
If a file in the destination needs to be modified, the symlink should be
deleted and replaced with an actual file;
this avoids mutating the cache and it ensures that
:py:func:`~gypsum_client.prepare_directory_for_upload.prepare_directory_upload`
recognizes that a new file actually needs to be uploaded.

Advanced users can set ``download=False``, in which case symlinks are created
even if their targets are not present in the cache.
In such cases, the destination should be treated as write-only due to the
potential presence of dangling symlinks.
This mode is useful for uploading a new version of an asset without
downloading the files from the existing version,
assuming that the modifications associated with the former can be
achieved without reading any of the latter.

On Windows, the user may not have permissions to create symbolic links,
so the function will transparently fall back to creating hard links or
copies instead.
This precludes any optimization by prepare_directory_upload as the hard
links/copies cannot be converted into upload links.
It also assumes that download=True as dangling links/copies cannot be created.
"""

import errno
import os
import shutil

from ._utils import BUCKET_CACHE_NAME
from .cache_directory import cache_directory
from .fetch_operations import fetch_manifest
from .rest_url import rest_url
from .save_operations import save_version

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def clone_version(
    project: str,
    asset: str,
    version: str,
    destination: str,
    download: bool = True,
    cache_dir: str = cache_directory(),
    url: str = rest_url(),
    **kwargs,
):
    """Clone a version's directory structure.

    Clone the directory structure for a versioned asset into a separate location.
    This is typically used to prepare a new version for a lightweight upload.

    See Also:
        :py:func:`~gypsum_client.prepare_directory_for_upload.prepare_directory_upload`,
        to prepare an upload based on the directory contents.

    Example:

        .. code-block:: python

            import tempfile

            cache = tempfile.mkdtemp()
            dest = tempfile.mkdtemp()

            clone_version("test-R", "basic", "v1", destination=dest, cache_dir=cache)

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        cache_dir:
            Path to the cache directory.

        destination:
            Destination directory at which to create the clone.

        download:
            Whether the version's files should be downloaded first.
            This can be set to `False` to create a clone without
            actually downloading any of the version's files.
            Defaults to True.

        url:
            URL of the gypsum REST API.

        **kwargs:
            Further arguments to pass to
            :py:func:`~gypsum_client.save_assets.save_version`.

            Only used if ``download`` is `True`.
    """
    if download:
        save_version(project, asset, version, cache_dir=cache_dir, url=url, **kwargs)

    final_cache = os.path.join(cache_dir, BUCKET_CACHE_NAME, project, asset, version)
    listing = fetch_manifest(project, asset, version, cache_dir=cache_dir, url=url)
    os.makedirs(destination, exist_ok=True)

    # Normalize final_cache path
    final_cache = os.path.abspath(final_cache)

    # Create symlinks back to the cache
    for file_name in listing.keys():
        dpath = os.path.join(destination, file_name)
        os.makedirs(os.path.dirname(dpath), exist_ok=True)
        target = os.path.join(final_cache, file_name)

        try:
            os.symlink(target, dpath)
        except OSError as e:
            if e.errno == errno.EEXIST:
                os.remove(dpath)
                os.symlink(target, dpath)
            elif os.name == "nt":
                try:
                    os.link(target, dpath)
                except OSError:
                    shutil.copy(target, dpath)
            else:
                raise RuntimeError(
                    f"failed to create a symbolic link to '{target}' at '{dpath}'."
                ) from e



from gypsum_client._utils import _remove_slash_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_truncating_url():
    single_slash = _remove_slash_url("https://jkanche.com/")
    assert single_slash == "https://jkanche.com"

    double_slash = _remove_slash_url("https://jkanche.com//")
    assert double_slash == "https://jkanche.com"



"""
Set this to False if SSL certificates are not properly setup on your machine.

essentially sets ``verify=False`` on all requests to the gypsum REST API.

Example:

    .. code-block::python

        from gypsum_client import REQUESTS_MOD
        # to set verify to False
        REQUESTS_MOD["verify"] = False

"""

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

REQUESTS_MOD = {"verify": True}



import os
import tempfile

import pytest
from gypsum_client import clone_version

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_clone_version_works_as_expected_with_existing_files():
    cache = tempfile.mkdtemp()
    dest = tempfile.mkdtemp()
    clone_version("test-R", "basic", "v1", destination=dest, cache_dir=cache)

    d1 = os.path.join(dest, "blah.txt")
    if os.name != "nt":  # os.name != 'nt' for non-Windows platforms
        l1 = os.readlink(d1)
        assert os.path.exists(l1)
        assert l1.endswith("test-R/basic/v1/blah.txt")
    else:
        assert os.path.exists(d1)

    d2 = os.path.join(dest, "foo/bar.txt")
    if os.name != "nt":
        l2 = os.readlink(d2)
        assert os.path.exists(l2)
        assert l2.endswith("test-R/basic/v1/foo/bar.txt")
    else:
        assert os.path.exists(d2)


@pytest.mark.skipif(
    os.name == "nt",
    reason="download=False can't work on Windows if symbolic links aren't available.",
)
def test_clone_version_happily_works_without_any_download():
    cache = tempfile.mkdtemp()
    dest = tempfile.mkdtemp()
    clone_version(
        "test-R", "basic", "v1", download=False, destination=dest, cache_dir=cache
    )

    l1 = os.readlink(os.path.join(dest, "blah.txt"))
    assert l1.endswith("test-R/basic/v1/blah.txt")
    assert not os.path.exists(l1)

    l2 = os.readlink(os.path.join(dest, "foo/bar.txt"))
    assert l2.endswith("test-R/basic/v1/foo/bar.txt")
    assert not os.path.exists(l2)


@pytest.mark.skipif(
    os.name == "nt",
    reason="download=False can't work on Windows if symbolic links aren't available.",
)
def test_clone_version_works_correctly_if_version_itself_contains_links():
    cache = tempfile.mkdtemp()
    dest = tempfile.mkdtemp()
    clone_version(
        "test-R",
        "basic",
        "v2",
        destination=dest,
        cache_dir=cache,
        download=False,
        relink=False,
    )

    l1 = os.readlink(os.path.join(dest, "blah.txt"))
    assert l1.endswith("test-R/basic/v2/blah.txt")
    assert not os.path.exists(l1)

    l2 = os.readlink(os.path.join(dest, "foo/bar.txt"))
    assert l2.endswith("test-R/basic/v2/foo/bar.txt")
    assert not os.path.exists(l2)



import os
import tempfile

import pytest
from gypsum_client import save_file, save_version

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

blah_contents = (
    "A\nB\nC\nD\nE\nF\nG\nH\nI\nJ\nK\nL\nM\nN\nO\nP\nQ\nR\nS\nT\nU\nV\nW\nX\nY\nZ\n"
)
foobar_contents = "1 2 3 4 5\n6 7 8 9 10\n"


def test_save_file_works_as_expected():
    cache = tempfile.mkdtemp()

    out = save_file("test-R", "basic", "v1", "blah.txt", cache_dir=cache)
    assert open(out, "r").read() == blah_contents

    out = save_file("test-R", "basic", "v1", "foo/bar.txt", cache_dir=cache)
    assert open(out, "r").read() == foobar_contents

    # Actually uses the cache.
    with open(out, "w") as f:
        f.write("".join(list(map(chr, range(97, 123)))))
    out = save_file("test-R", "basic", "v1", "foo/bar.txt", cache_dir=cache)
    assert open(out, "r").read() == "".join(list(map(chr, range(97, 123))))

    # Until it doesn't.
    out = save_file(
        "test-R", "basic", "v1", "foo/bar.txt", cache_dir=cache, overwrite=True
    )
    assert open(out, "r").read() == foobar_contents

    with pytest.raises(Exception):
        save_file("test-R", "basic", "v1", "no.exist.txt", cache_dir=cache)


def test_save_file_works_via_links():
    cache = tempfile.mkdtemp()
    out = save_file("test-R", "basic", "v2", "blah.txt", cache_dir=cache)
    assert open(out, "r").read() == blah_contents

    out = save_file("test-R", "basic", "v2", "foo/bar.txt", cache_dir=cache)
    assert open(out, "r").read() == foobar_contents

    # Pulling out some ancestors.
    out = save_file("test-R", "basic", "v3", "blah.txt", cache_dir=cache)
    assert open(out, "r").read() == blah_contents

    out = save_file("test-R", "basic", "v3", "foo/bar.txt", cache_dir=cache)
    assert open(out, "r").read() == foobar_contents


def test_save_version_works_as_expected_without_links():
    cache = tempfile.mkdtemp()

    out = save_version("test-R", "basic", "v1", cache_dir=cache)
    assert os.path.exists(os.path.join(out, "..manifest"))
    assert open(os.path.join(out, "blah.txt"), "r").read() == blah_contents
    assert open(os.path.join(out, "foo", "bar.txt"), "r").read() == foobar_contents

    # Caching behaves as expected.
    path = os.path.join(out, "blah.txt")
    with open(path, "w") as f:
        f.write("foo")
    out = save_version("test-R", "basic", "v1", cache_dir=cache)
    assert open(path, "r").read() == "foo"

    # Unless we overwrite.
    out = save_version("test-R", "basic", "v1", cache_dir=cache, overwrite=True)
    assert open(path, "r").read() == blah_contents

    # Also works concurrently.
    cache2 = tempfile.mkdtemp()

    out = save_version("test-R", "basic", "v1", cache_dir=cache2, concurrent=2)
    assert os.path.exists(
        os.path.join(cache2, "bucket", "test-R", "basic", "v1", "..manifest")
    )
    assert (
        open(
            os.path.join(cache2, "bucket", "test-R", "basic", "v1", "blah.txt"), "r"
        ).read()
        == blah_contents
    )
    assert (
        open(
            os.path.join(cache2, "bucket", "test-R", "basic", "v1", "foo", "bar.txt"),
            "r",
        ).read()
        == foobar_contents
    )


def test_save_version_works_as_expected_with_links():
    cache = tempfile.mkdtemp()

    out = save_version("test-R", "basic", "v2", cache_dir=cache)
    assert open(os.path.join(out, "blah.txt"), "r").read() == blah_contents
    assert open(os.path.join(out, "foo", "bar.txt"), "r").read() == foobar_contents
    assert os.path.exists(
        os.path.join(cache, "bucket", "test-R", "basic", "v1", "blah.txt")
    )  # populates the linked-to files as well.
    assert os.path.exists(
        os.path.join(cache, "bucket", "test-R", "basic", "v1", "foo", "bar.txt")
    )

    # Unless we turn off link resolution.
    cache = tempfile.mkdtemp()

    out = save_version("test-R", "basic", "v2", cache_dir=cache, relink=False)
    assert not os.path.exists(os.path.join(out, "blah.txt"))
    assert not os.path.exists(os.path.join(out, "foo", "bar.txt"))

    # Works recursively.
    cache = tempfile.mkdtemp()

    out = save_version("test-R", "basic", "v3", cache_dir=cache)
    assert open(os.path.join(out, "blah.txt"), "r").read() == blah_contents
    assert open(os.path.join(out, "foo", "bar.txt"), "r").read() == foobar_contents
    assert os.path.exists(
        os.path.join(cache, "bucket", "test-R", "basic", "v1", "blah.txt")
    )  # populates the ancestral versions as well.
    assert os.path.exists(
        os.path.join(cache, "bucket", "test-R", "basic", "v1", "foo", "bar.txt")
    )

    # Link resolver doesn't do more work if it's already present.
    path = os.path.join(out, "foo", "bar.txt")
    with open(path, "w") as f:
        f.write("Aaron wuz here")
    out = save_version("test-R", "basic", "v3", cache_dir=cache)
    assert open(path, "r").read() == "Aaron wuz here"

    # Unless we force it to.
    out = save_version("test-R", "basic", "v3", cache_dir=cache, overwrite=True)
    assert open(path, "r").read() == foobar_contents



import os

import pytest
from gypsum_client import fetch_latest, refresh_latest

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

app_url = "https://gypsum-test.artifactdb.com"


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_refresh_latest():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    ver = refresh_latest("test-R", "basic", url=app_url, token=gh_token)
    assert ver == "v3"
    assert fetch_latest("test-R", "basic", url=app_url) == ver



import os
import tempfile
from datetime import datetime, timedelta

import pytest
from gypsum_client.auth import access_token, set_access_token

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_setting_the_token_works_as_expected():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    out = set_access_token(gh_token, cache_dir=None)
    assert out["token"] == gh_token
    assert access_token(cache_dir=None) == gh_token

    # Works when we cache it to disk.
    with tempfile.TemporaryDirectory() as cache:
        out = set_access_token(gh_token, cache_dir=cache)
        assert out["token"] == gh_token
        assert access_token(cache_dir=cache) == gh_token

        tokpath = os.path.join(cache, "credentials", "token.txt")
        with open(tokpath, "r") as f:
            assert f.readline().strip() == gh_token

        # Unsetting it works as expected.
        out = set_access_token(None, cache_dir=None)
        # Assuming gypsum.token_cache.auth_info should be None
        assert getattr(set_access_token, "auth_info", None) is None
        assert os.path.exists(tokpath)

        out = set_access_token(None, cache_dir=cache)
        assert getattr(set_access_token, "auth_info", None) is None
        assert not os.path.exists(tokpath)


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_access_token_retrieval_works_as_expected():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    set_access_token(None, cache_dir=None)  # wipe out in-memory cache.

    with tempfile.TemporaryDirectory() as cache:
        tokpath = os.path.join(cache, "credentials", "token.txt")
        os.makedirs(os.path.dirname(tokpath), exist_ok=True)
        with open(tokpath, "w") as f:
            f.write(
                f"foobar\nurmom\n{(datetime.now() + timedelta(seconds=10000)).timestamp()}\n"
            )

        info = access_token(full=True, request=False, cache_dir=cache)
        assert info["token"] == "foobar"
        assert info["name"] == "urmom"
        assert isinstance(info["expires"], float)

        # Works from cache.
        info = access_token(full=True, request=False, cache_dir=None)
        assert info["token"] == "foobar"
        assert info["name"] == "urmom"
        assert isinstance(info["expires"], float)

        # Falls through to request.
        set_access_token(None, cache_dir=None)  # wipe out in-memory cache.
        os.unlink(tokpath)
        info = access_token(full=True, request=False, cache_dir=cache)
        assert info is None


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_access_token_expiry_works_as_expected():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")
    set_access_token(None, cache_dir=None)  # wipe out in-memory cache.

    with tempfile.TemporaryDirectory() as cache:
        tokpath = os.path.join(cache, "credentials", "token.txt")
        os.makedirs(os.path.dirname(tokpath), exist_ok=True)
        with open(tokpath, "w") as f:
            f.write(
                "foobar\nurmom\n{}\n".format(
                    (datetime.now() - timedelta(seconds=10000)).timestamp()
                )
            )

        info = access_token(full=True, request=False, cache_dir=cache)
        assert info is None

        # Works in memory as well.
        info = access_token(full=True, request=False, cache_dir=None)
        assert info is None



import os

import pytest
from gypsum_client import (
    approve_probation,
    complete_upload,
    fetch_summary,
    reject_probation,
    remove_asset,
    start_upload,
)

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

app_url = "https://gypsum-test.artifactdb.com"


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_probation_approve():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_asset("test-Py", "probation", url=app_url, token=gh_token)

    init = start_upload(
        project="test-Py",
        asset="probation",
        version="v1",
        files=[],
        probation=True,
        token=gh_token,
        url=app_url,
    )

    complete_upload(init, url=app_url)

    _summary = fetch_summary("test-Py", "probation", "v1", cache_dir=None, url=app_url)
    assert _summary["on_probation"] is True

    approve_probation("test-Py", "probation", "v1", token=gh_token, url=app_url)
    _summary_after = fetch_summary(
        "test-Py", "probation", "v1", cache_dir=None, url=app_url
    )
    assert "on_probation" not in _summary_after


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_probation_reject():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_asset("test-Py", "probation", url=app_url, token=gh_token)

    init = start_upload(
        project="test-Py",
        asset="probation",
        version="v1",
        files=[],
        probation=True,
        token=gh_token,
        url=app_url,
    )

    complete_upload(init, url=app_url)

    _summary = fetch_summary("test-Py", "probation", "v1", cache_dir=None, url=app_url)
    assert _summary["on_probation"] is True

    reject_probation("test-Py", "probation", "v1", token=gh_token, url=app_url)

    with pytest.raises(Exception):
        fetch_summary("test-Py", "probation", "v1", cache_dir=None, url=app_url)



import os
from datetime import datetime, timedelta, timezone

import pytest
from gypsum_client import (
    create_project,
    fetch_permissions,
    remove_project,
    set_permissions,
)

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

app_url = "https://gypsum-test.artifactdb.com"


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_permission_setting():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_project("test-Py-perms", token=gh_token, url=app_url)
    create_project("test-Py-perms", owners=["jkanche"], token=gh_token, url=app_url)

    until = (
        (datetime.now() + timedelta(seconds=1000000))
        .replace(microsecond=0)
        # .strftime("%Y-%m-%dT%H:%M:%S%Z")
    )

    set_permissions(
        "test-Py-perms",
        owners=["LTLA"],
        uploaders=[{"id": "lawremi", "until": until}],
        token=gh_token,
        url=app_url,
    )

    perms = fetch_permissions("test-Py-perms", url=app_url)
    assert sorted(perms["owners"]) == sorted(["LTLA", "jkanche"])
    assert len(perms["uploaders"]) == 1
    assert perms["uploaders"][0]["id"] == "lawremi"
    assert perms["uploaders"][0]["until"] == until

    # Checking uploader appending, while also checking owners=None.
    set_permissions(
        "test-Py-perms",
        uploaders=[{"id": "ArtifactDB-bot", "trusted": True}],
        token=gh_token,
        url=app_url,
    )
    perms = fetch_permissions("test-Py-perms", url=app_url)
    assert sorted(perms["owners"]) == sorted(["LTLA", "jkanche"])
    assert len(perms["uploaders"]) == 2
    assert perms["uploaders"][0]["id"] == "lawremi"
    assert perms["uploaders"][1]["id"] == "ArtifactDB-bot"
    assert perms["uploaders"][1]["trusted"] is True

    # Checking union of owners, and also that uploaders=None works.
    set_permissions(
        "test-Py-perms", owners=["PeteHaitch", "LTLA"], token=gh_token, url=app_url
    )
    perms = fetch_permissions("test-Py-perms", url=app_url)
    assert sorted(perms["owners"]) == sorted(["LTLA", "jkanche", "PeteHaitch"])
    assert len(perms["uploaders"]) == 2

    # Resetting the owners back.
    set_permissions(
        "test-Py-perms", owners=["jkanche"], append=False, token=gh_token, url=app_url
    )
    perms = fetch_permissions("test-Py-perms", url=app_url)
    assert perms["owners"] == ["jkanche"]
    assert len(perms["uploaders"]) == 2

    # Now resetting the uploaders.
    set_permissions(
        "test-Py-perms", uploaders=[], append=False, token=gh_token, url=app_url
    )
    perms = fetch_permissions("test-Py-perms", url=app_url)
    assert perms["owners"] == ["jkanche"]
    assert len(perms["uploaders"]) == 0



import json
import tempfile

import pytest
from gypsum_client.fetch_metadata_schema import fetch_metadata_schema
from gypsum_client.validate_metadata import validate_metadata

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_fetchMetadataSchema():
    _cache_dir = tempfile.mkdtemp()
    path = fetch_metadata_schema(cache_dir=_cache_dir)
    assert isinstance(json.load(open(path)), dict)

    # Uses the cache
    with open(path, "w") as f:
        f.write("FOO_BAR")

    with pytest.raises(Exception):
        json.load(open(path))

    path2 = fetch_metadata_schema(cache_dir=_cache_dir)
    assert path == path2
    assert open(path).read().strip() == "FOO_BAR"

    # Unless we overwrite it
    man = fetch_metadata_schema(cache_dir=_cache_dir, overwrite=True)
    assert isinstance(json.load(open(path)), dict)


def test_validateMetadata():

    _cache_dir = tempfile.mkdtemp()

    metadata = {
        "title": "Fatherhood",
        "description": "Luke ich bin dein Vater.",
        "sources": [{"provider": "GEO", "id": "GSE12345"}],
        "taxonomy_id": ["9606"],
        "genome": ["GRCm38"],
        "maintainer_name": "Darth Vader",
        "maintainer_email": "vader@empire.gov",
        "bioconductor_version": "3.10",
    }

    schema = fetch_metadata_schema(cache_dir=_cache_dir)
    assert validate_metadata(metadata, schema)

    assert validate_metadata(json.dumps(metadata), schema)

    metadata.pop("bioconductor_version", None)
    with pytest.raises(Exception):
        validate_metadata(metadata, schema)



"""Prepare to upload a directory's contents.

Files in `directory` (that are not symlinks) are used as
regular uploads, i.e., `files=` in
:py:func:`~gypsum_client.upload_api_operations.start_upload`.

If `directory` contains a symlink to a file in `cache`,
we assume that it points to a file that was previously downloaded
by, e.g., :py:func:`~gypsum_client.upload_api_operations.save_file` or
:py:func:`~gypsum_client.upload_api_operations.save_version`.
Thus, instead of performing a regular upload, we attempt to
create an upload link, i.e., ``links=`` in
:py:func:`~gypsum_client.upload_api_operations.start_upload`.
This is achieved by examining the destination path of the
symlink and inferring the link destination in the backend.
Note that this still works if the symlinks are dangling.

If a symlink cannot be converted into an upload link, it will
be used as a regular upload, i.e., the contents of the symlink
destination will be uploaded by
:py:func:`~gypsum_client.upload_api_operations.start_upload`.
In this case, an error will be raised if the symlink is dangling
as there is no file that can actually be uploaded.
If ``links="always"``, an error is raised instead upon symlink
conversion failure.

This function is intended to be used with
:py:func:`~gypsum_client.clone_operations.clone_version`,
which creates symlinks to files in `cache`.

See Also:
    :py:func:`~gypsum_client.upload_api_operations.start_upload`,
    to actually start the upload.

    :py:func:`~gypsum_client.clone_operations.clone_version`,
    to prepare the symlinks.

Example:

    .. code-block:: python

        import tempfile
        cache = tempfile.mkdtemp()
        dest = tempfile.mkdtemp()

        # Clone a project
        clone_version("test-R", "basic", "v1", destination=dest, cache_dir=cache)

        # Make some modification
        with open(os.path.join(dest, "heanna"), "w") as f:
            f.write("sumire")

        # Prepare the directory for upload
        prepped = prepare_directory_upload(dest, cache_dir=cache)
"""

import os
from typing import Literal

from ._utils import (
    BUCKET_CACHE_NAME,
    _sanitize_path,
)
from .cache_directory import cache_directory

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def prepare_directory_upload(
    directory: str,
    links: Literal["auto", "always", "never"] = "auto",
    cache_dir: str = cache_directory(),
) -> dict:
    """Prepare to upload a directory's contents.

    Prepare to upload a directory's contents via `start_upload`.
    This goes through the directory to list its contents and
    convert symlinks to upload links.

    Args:
        directory:
            Path to a directory, the contents of which are to be
            uploaded via :py:func:`~gypsum_client.start_upload.start_upload`.

        links:
            Indicate how to handle symlinks in `directory`.
            Must be one of the following:
            - "auto": Will attempt to convert symlinks into upload links.
            If the conversion fails, a regular upload is performed.
            - "always": Will attempt to convert symlinks into upload links.
            If the conversion fails, an error is raised.
            - "never": Will never attempt to convert symlinks into upload
            links. All symlinked files are treated as regular uploads.

        cache_dir:
            Path to the cache directory, used to convert symlinks into upload links.

    Returns:
        Dictionary containing:
        - `files`: list of strings to be used as `files=`
        in :py:func:`~gypsum_client.start_upload.start_upload`.
        - `links`: dictionary to be used as `links=` in
        :py:func:`~gypsum_client.start_upload.start_upload`.

    """
    _links_options = ["auto", "always", "never"]
    if links not in _links_options:
        raise ValueError(
            f"Invalid value for 'links': {links}. Must be one of {_links_options}."
        )

    out_files = []
    out_links = []

    cache_dir = _normalize_and_sanitize_path(cache_dir)
    if not cache_dir.endswith("/"):
        cache_dir += "/"

    for root, _, files in os.walk(directory):
        for name in files:
            rel_path = os.path.relpath(os.path.join(root, name), directory)

            if not os.path.islink(os.path.join(directory, rel_path)):
                out_files.append(rel_path)
                continue

            dest = os.readlink(os.path.join(directory, rel_path))

            if links == "never":
                if not os.path.exists(dest):
                    raise ValueError(
                        f"Cannot use a dangling link to '{dest}' as a regular upload."
                    )
                out_files.append(rel_path)
                continue

            dest = _normalize_and_sanitize_path(dest)
            dest_components = _match_path_to_cache(dest, cache_dir)

            if dest_components:
                out_links.append(
                    {
                        "from.path": rel_path,
                        "to.project": dest_components["project"],
                        "to.asset": dest_components["asset"],
                        "to.version": dest_components["version"],
                        "to.path": dest_components["path"],
                    }
                )
                continue

            if links == "always":
                raise ValueError(
                    f"Failed to convert symlink '{dest}' to an upload link."
                )
            elif not os.path.exists(dest):
                raise ValueError(
                    f"Cannot use a dangling link to '{dest}' as a regular upload."
                )

            out_files.append(rel_path)

    return {"files": out_files, "links": out_links}


def _normalize_and_sanitize_path(path: str) -> str:
    if os.path.exists(path):
        path = os.path.join(
            os.path.normpath(os.path.dirname(path)), os.path.basename(path)
        )
    return _sanitize_path(path)


def _match_path_to_cache(path: str, cache: str) -> dict:
    if not path.startswith(cache):
        return None

    remainder = path[len(cache) :]
    components = remainder.split("/")
    if len(components) <= 4 or components[0] != BUCKET_CACHE_NAME:
        return None

    return {
        "project": components[1],
        "asset": components[2],
        "version": components[3],
        "path": "/".join(components[4:]),
    }



import os
import shutil
import tempfile

import pytest
from gypsum_client import clone_version, prepare_directory_upload

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_prepare_directory_upload_works_as_expected():
    cache = tempfile.mkdtemp()
    dest = tempfile.mkdtemp()
    clone_version("test-R", "basic", "v1", destination=dest, cache_dir=cache)
    with open(os.path.join(dest, "heanna"), "w") as f:
        f.write("sumire")

    prepped = prepare_directory_upload(dest, cache_dir=cache)
    if os.name == "nt":  # Windows
        assert prepped["files"] == ["blah.txt", "foo/bar.txt", "heanna"]
        assert len(prepped["links"]) == 0
    else:  # Unix
        assert prepped["files"] == ["heanna"]
        assert sorted(x["from.path"] for x in prepped["links"]) == sorted(
            ["blah.txt", "foo/bar.txt"]
        )
        assert sorted(x["to.project"] for x in prepped["links"]) == sorted(
            ["test-R", "test-R"]
        )
        assert sorted(x["to.asset"] for x in prepped["links"]) == sorted(
            ["basic", "basic"]
        )
        assert sorted(x["to.version"] for x in prepped["links"]) == sorted(["v1", "v1"])
        assert sorted(x["to.path"] for x in prepped["links"]) == sorted(
            ["blah.txt", "foo/bar.txt"]
        )

    prepped = prepare_directory_upload(dest, cache_dir=cache, links="never")
    assert sorted(prepped["files"]) == sorted(["blah.txt", "foo/bar.txt", "heanna"])
    assert len(prepped["links"]) == 0

    shutil.rmtree(cache)
    shutil.rmtree(dest)


@pytest.mark.skipif(os.name == "nt", reason="skipping symlink tests on Windows")
def test_prepare_directory_upload_with_some_odd_things():
    cache = tempfile.mkdtemp()
    dest = tempfile.mkdtemp()
    clone_version("test-R", "basic", "v1", destination=dest, cache_dir=cache)

    with open(os.path.join(dest, "..check"), "w") as f:
        f.write("stuff")

    random = tempfile.mktemp()
    os.symlink(random, os.path.join(dest, "arkansas"))

    with pytest.raises(Exception):
        prepare_directory_upload(dest, cache_dir=cache, links="always")

    with pytest.raises(Exception):
        prepare_directory_upload(dest, cache_dir=cache, links="never")

    with pytest.raises(Exception):
        prepare_directory_upload(dest, cache_dir=cache)

    with open(random, "w") as f:
        f.write("YAY")

    prepped = prepare_directory_upload(dest, cache_dir=cache)
    assert sorted(prepped["files"]) == sorted(["..check", "arkansas"])
    assert sorted(x["from.path"] for x in prepped["links"]) == sorted(
        ["blah.txt", "foo/bar.txt"]
    )

    shutil.rmtree(cache)
    shutil.rmtree(dest)


@pytest.mark.skipif(os.name == "nt", reason="skipping symlink tests on Windows")
def test_prepare_directory_upload_handles_dangling_links_correctly():
    cache = tempfile.mkdtemp()
    dest = tempfile.mkdtemp()
    clone_version(
        "test-R", "basic", "v1", destination=dest, cache_dir=cache, download=False
    )

    prepped = prepare_directory_upload(dest, cache_dir=cache)
    assert prepped["files"] == []
    assert len(prepped["links"]) == 2

    with pytest.raises(Exception):
        prepare_directory_upload(dest, cache_dir=cache, links="never")

    shutil.rmtree(cache)
    shutil.rmtree(dest)



import json
import os
import tempfile
from datetime import datetime

import pytest
from gypsum_client.fetch_operations import (
    fetch_latest,
    fetch_manifest,
    fetch_permissions,
    fetch_quota,
    fetch_summary,
    fetch_usage,
)
from gypsum_client.fetch_metadata_schema import fetch_metadata_schema

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_fetch_manifest():
    cache = tempfile.mkdtemp()
    man = fetch_manifest("test-R", "basic", "v1", cache_dir=cache)
    assert sorted(man.keys()) == ["blah.txt", "foo/bar.txt"]

    # Uses the cache.
    with open(
        os.path.join(cache, "bucket", "test-R", "basic", "v1", "..manifest"), "w"
    ) as f:
        f.write("[]")

    man = fetch_manifest("test-R", "basic", "v1", cache_dir=cache)
    assert len(man) == 0

    # Unless we overwrite it.
    man = fetch_manifest("test-R", "basic", "v1", cache_dir=cache, overwrite=True)
    assert len(man) > 0

    with pytest.raises(Exception):
        fetch_manifest("test-R", "basic", "non-existent", cache_dir=cache)


def test_fetch_summary():
    cache = tempfile.mkdtemp()
    xx = fetch_summary("test-R", "basic", "v1", cache_dir=cache)
    original_user = xx["upload_user_id"]
    assert isinstance(xx["upload_start"], datetime)
    assert isinstance(xx["upload_finish"], datetime)

    # Uses the cache.
    sumpath = os.path.join(cache, "bucket", "test-R", "basic", "v1", "..summary")
    with open(sumpath, "w") as f:
        json.dump(
            {
                "upload_user_id": "adrian",
                "upload_start": "2022-01-01T01:01:01Z",
                "upload_finish": "2022-01-01T01:01:02Z",
            },
            f,
        )
    xx = fetch_summary("test-R", "basic", "v1", cache_dir=cache)
    assert xx["upload_user_id"] == "adrian"

    # Unless we overwrite it.
    xx = fetch_summary("test-R", "basic", "v1", cache_dir=cache, overwrite=True)
    assert xx["upload_user_id"] == original_user

    # Self-deletes from the cache if it's on probation.
    with open(sumpath, "w") as f:
        json.dump(
            {
                "upload_user_id": "adrian",
                "upload_start": "2022-01-01T01:01:01Z",
                "upload_finish": "2022-01-01T01:01:02Z",
                "on_probation": True,
            },
            f,
        )
    xx = fetch_summary("test-R", "basic", "v1", cache_dir=cache)
    assert xx["on_probation"]
    assert not os.path.exists(sumpath)

    with pytest.raises(Exception):
        fetch_summary("test-R", "basic", "non-existent", cache_dir=cache)


def test_fetch_latest():
    assert fetch_latest("test-R", "basic") == "v3"


def test_fetch_usage():
    assert fetch_usage("test-R") > 0


def test_fetch_permissions():
    perms = fetch_permissions("test-R")
    assert isinstance(perms["owners"], list)
    assert isinstance(perms["uploaders"], list)



from urllib.parse import quote_plus

import requests

from ._utils import _remove_slash_url
from .auth import access_token
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def remove_asset(project: str, asset: str, url: str = rest_url(), token: str = None):
    """Remove an asset of a project from the gypsum backend.

    See Also:
        :py:func:`~.remove_project`,
        to remove a project.

        :py:func:`~.remove_version`,
        to remove a specific version.

    Example:

        .. code-block:: python

            # Mock a project
            init = start_upload(
                project="test-Py-remove",
                asset="mock-remove",
                version="v1",
                files=[],
            )

            complete_upload(init)
            remove_asset("test-Py-remove", "mock-remove")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.
            The token must refer to a gypsum administrator account.

    Returns:
        True if asset was successfully removed.
    """

    if token is None:
        token = access_token()

    _key = f"{quote_plus(project)}/{quote_plus(asset)}"
    _request_removal(_key, url=url, token=token)

    return True


def remove_project(project: str, url: str = rest_url(), token: str = None):
    """Remove a project from the gypsum backend.

    See Also:
        :py:func:`~gypsum_client.create_operations.create_project`,
        to create a project.

        :py:func:`~.remove_asset`,
        to remove a specific asset.

        :py:func:`~.remove_version`,
        to remove a specific version.

    Example:

        .. code-block:: python

            create_project("test-Py-remove", owners=["jkanche"])
            remove_project("test-Py-remove")

    Args:
        project:
            Project name.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.
            The token must refer to a gypsum administrator account.

    Returns:
        True if the project was successfully removed.
    """
    if token is None:
        token = access_token()

    _key = f"{quote_plus(project)}"
    _request_removal(_key, url=url, token=token)

    return True


def remove_version(
    project: str,
    asset: str,
    version: str,
    url: str = rest_url(),
    token: str = None,
):
    """Remove a project from the gypsum backend.

    See Also:
        :py:func:`~.remove_asset`,
        to remove a specific asset.

        :py:func:`~.remove_version`,
        to remove a specific version.

    Example:

        .. code-block:: python

            # Mock a project
            init = start_upload(
                project="test-Py-remove",
                asset="mock-remove",
                version="v1",
                files=[],
            )

            complete_upload(init)

            remove_version("test-Py-remove", "mock-remove", "v1")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.

    Returns:
        True if the version of the project was successfully removed.
    """

    if token is None:
        token = access_token()

    _key = f"{quote_plus(project)}/{quote_plus(asset)}/{quote_plus(version)}"
    _request_removal(_key, url=url, token=token)

    return True


def _request_removal(suffix: str, url: str, token: str):
    url = _remove_slash_url(url)

    headers = {}
    headers["Authorization"] = f"Bearer {token}"

    req = requests.delete(f"{url}/remove/{suffix}", headers=headers)
    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to remove assets in the project, {req.status_code} and reason: {req.text}"
        ) from e

    return True



import hashlib
import os
from typing import List, Union
from urllib.parse import quote_plus

import requests

from ._utils import _remove_slash_url, _sanitize_path
from .auth import access_token
from .config import REQUESTS_MOD
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def start_upload(
    project: str,
    asset: str,
    version: str,
    files: Union[str, List[str], List[dict]],
    links: List[dict] = None,
    deduplicate: bool = True,
    probation: bool = False,
    url: str = rest_url(),
    token: str = None,
    directory: str = None,
) -> dict:
    """Start an upload.

    Start an upload of a new version of an asset,
    or a new asset of a project.

    See Also:
        :py:func:`~gypsum_client.upload_file_operations.upload_files`,
        to actually upload the files.

        :py:func:`~.complete_upload`,
        to indicate that the upload is completed.

        :py:func:`~.abort_upload`,
        to abort an upload in progress.

        :py:func:`~gypsum_client.prepare_directory_for_upload.prepare_directory_upload`,
        to create ``files`` and ``links`` from a directory.

    Example:

        .. code-block:: python

            import tempfile
            tmp_dir = tempfile.mkdtemp()

            with open(f"{tmp_dir}/blah.txt", "w") as f:
                f.write(blah_contents)

            os.makedirs(f"{tmp_dir}/foo", exist_ok=True)

            with open(f"{tmp_dir}/foo/blah.txt", "w") as f:
                f.write(foobar_contents)

            files = [
                str(file.relative_to(tmp_dir))
                for file in Path(tmp_dir).rglob("*")
                if not os.path.isdir(file)
            ]

            init = start_upload(
                project="test-Py-demo",
                asset="upload",
                version="1",
                files=files,
                directory=tmp_dir,
            )

            abort_upload(init)

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        files:
            A file path or a List of file paths to upload.
            These paths are assumed to be relative to the
            ``directory`` parameter.

            Optionally, May be provided a list where each element
            is a dictionary containing the following keys:
            - ``path``: a string containing the relative path of the
            file inside the version's subdirectory.
            - ``size``, a non-negative integer specifying the size of the
            file in bytes.
            - ``md5sum``, a string containing the hex-encoded MD5
            checksum of the file.
            - Optionally ``dedup``, a boolean value indicating
            whether deduplication should be attempted for each file. If this is
            not available, the parameter ``deduplicate`` is used.


        links:
            A List containing a dictionary with the following keys:
            - ``from.path``: a string containing the relative path of the
            file inside the version's subdirectory.
            - ``to.project``: a string containing the project of the list
            destination.
            - ``to.asset``: a string containing the asset of the list
            destination.
            - ``to.version``: a string containing the version of the list
            destination.
            - ``to.path``: a string containing the path of the list destination.

        deduplicate:
            Whether the backend should attempt deduplication of ``files``
            in the immediately previous version.
            Defaults to True.

        probation:
            Whether to perform a probational upload.
            Defaults to False.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.

        directory:
            Path to a directory containing the ``files`` to be uploaded.
            This directory is assumed to correspond to a version of an asset.

    Returns:
        Dictionary containing the following keys:
        - ``file_urls``, a list of lists containing information about each
        file to be uploaded. This is used by ``uploadFiles``.
        - ``complete_url``, a string containing the completion URL, to be
        used by ``complete_upload``.
        - ``abort_url``, a string specifying the abort URL, to be used by
        ``abort_upload``.
        - ``session_token``, a string for authenticating to the newly
        initialized upload session.
    """
    if isinstance(files, str):
        files = [files]

    _types_in_file = [type(f) for f in files]

    if len(set(_types_in_file)) > 1:
        raise ValueError(
            "All elements in 'files' must be strings or dicts, but not a mix."
        )

    _all_dict_in_files = all(isinstance(f, dict) for f in files)
    if _all_dict_in_files is False:
        _targets = []
        for f in files:
            if directory is not None:
                _targets.append(os.path.join(directory, f))
            else:
                _targets.append(f)

        _files_info = []
        for _tidx, _tg in enumerate(_targets):
            file_info = {
                "path": files[_tidx],
                "size": os.path.getsize(_tg),
                "md5sum": hashlib.md5(open(_tg, "rb").read()).hexdigest(),
                "dedup": deduplicate,
            }
            _files_info.append(file_info)
    else:
        _files_info = []
        for f in files:
            file_info = {
                "path": f["path"],
                "size": f["size"],
                "md5sum": f["md5sum"],
                "dedup": f["dedup"] if "dedup" in f else deduplicate,
            }
            _files_info.append(file_info)

    formatted = []
    for _, file in enumerate(_files_info):
        file_type = "simple" if file["dedup"] else "dedup"
        formatted.append(
            {
                "type": file_type,
                "path": _sanitize_path(file["path"]),
                "size": file["size"],
                "md5sum": file["md5sum"],
            }
        )

    if links is not None:
        out_links = []
        for _, link in enumerate(links):
            out_links.append(
                {
                    "type": "link",
                    "path": _sanitize_path(link["from.path"]),
                    "link": {
                        "project": link["to.project"],
                        "asset": link["to.asset"],
                        "version": link["to.version"],
                        "path": _sanitize_path(link["to.path"]),
                    },
                }
            )
        formatted.extend(out_links)

    if token is None:
        token = access_token()

    url = _remove_slash_url(url)
    req = requests.post(
        f"{url}/upload/start/{quote_plus(project)}/{quote_plus(asset)}/{quote_plus(version)}",
        json={"files": formatted, "on_probation": probation},
        headers={"Authorization": f"Bearer {token}"},
        verify=REQUESTS_MOD["verify"],
    )
    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to start an upload, {req.status_code} and reason: {req.text}"
        ) from e

    resp = req.json()

    if "status" in resp and resp["status"] == "error":
        raise Exception(
            f"Failed to upload, {req.status_code} and reason: {resp['reason']}"
        )

    return resp


def complete_upload(init: dict, url=rest_url()):
    """Complete an upload session after all files have been uploaded.

    See Also:
        :py:func:`~gypsum_client.upload_api_operations.start_upload`,
        to create the init.

    Example:

        .. code-block:: python

            import tempfile
            tmp_dir = tempfile.mkdtemp()

            with open(f"{tmp_dir}/blah.txt", "w") as f:
                f.write(blah_contents)

            os.makedirs(f"{tmp_dir}/foo", exist_ok=True)

            with open(f"{tmp_dir}/foo/blah.txt", "w") as f:
                f.write(foobar_contents)

            files = [
                str(file.relative_to(tmp_dir))
                for file in Path(tmp_dir).rglob("*")
                if not os.path.isdir(file)
            ]

            init = start_upload(
                project="test-Py-demo",
                asset="upload",
                version="1",
                files=files,
                directory=tmp_dir,
            )

            abort_upload(init)

    Args:
        init:
            Dictionary containing ``complete_url`` and ``session_token``.

            :py:func:`~.start_upload`, to create ``init``.

        url:
            URL to the gypsum REST API.
    """
    url = _remove_slash_url(url)
    req = requests.post(
        f"{url}{init['complete_url']}",
        headers={"Authorization": f"Bearer {init['session_token']}"},
    )
    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to complete an upload session, {req.status_code} and reason: {req.text}"
        ) from e


def abort_upload(init: dict, url=rest_url()):
    """Abort an upload session, usually after an irrecoverable error.

    See Also:
        :py:func:`~gypsum_client.upload_api_operations.start_upload`,
        to create the init.

    Example:

        .. code-block:: python

            import tempfile
            tmp_dir = tempfile.mkdtemp()

            with open(f"{tmp_dir}/blah.txt", "w") as f:
                f.write(blah_contents)

            os.makedirs(f"{tmp_dir}/foo", exist_ok=True)

            with open(f"{tmp_dir}/foo/blah.txt", "w") as f:
                f.write(foobar_contents)

            files = [
                str(file.relative_to(tmp_dir))
                for file in Path(tmp_dir).rglob("*")
                if not os.path.isdir(file)
            ]

            init = start_upload(
                project="test-Py-demo",
                asset="upload",
                version="1",
                files=files,
                directory=tmp_dir,
            )

            complete_upload(init)

    Args:
        init:
            Dictionary containing ``abort_url`` and ``session_token``.

            :py:func:`~.start_upload`, to create ``init``.

        url:
            URL to the gypsum REST API.
    """
    url = _remove_slash_url(url)
    req = requests.post(
        f"{url}{init['abort_url']}",
        headers={"Authorization": f"Bearer {init['session_token']}"},
    )

    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to abort the upload, {req.status_code} and reason: {req.text}"
        ) from e



import os
import tempfile
from pathlib import Path

import pytest
from gypsum_client import (
    create_project,
    fetch_quota,
    fetch_usage,
    refresh_usage,
    remove_project,
    set_quota,
)

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

blah_contents = (
    "A\nB\nC\nD\nE\nF\nG\nH\nI\nJ\nK\nL\nM\nN\nO\nP\nQ\nR\nS\nT\nU\nV\nW\nX\nY\nZ\n"
)
foobar_contents = "1 2 3 4 5\n6 7 8 9 10\n"

app_url = "https://gypsum-test.artifactdb.com"


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_quota():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_project("test-Py-quota", token=gh_token, url=app_url)
    create_project("test-Py-quota", owners=["jkanche"], token=gh_token, url=app_url)

    set_quota(
        "test-Py-quota",
        baseline=1234,
        growth_rate=5678,
        year=2020,
        token=gh_token,
        url=app_url,
    )

    quot = fetch_quota("test-Py-quota", url=app_url)
    assert quot["baseline"] == 1234
    assert quot["growth_rate"] == 5678
    assert quot["year"] == 2020


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_usage():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_project("test-Py-quota", token=gh_token, url=app_url)
    create_project("test-Py-quota", owners=["jkanche"], token=gh_token, url=app_url)

    # Fetch the initial usage
    initial_usage = fetch_usage("test-Py-quota", url=app_url)
    assert initial_usage == 0

    # Refresh the usage and check
    refreshed_usage = refresh_usage("test-Py-quota", url=app_url)
    assert refreshed_usage == 0



"""Text search on the metadata database.

Each string is tokenized by converting it to lower case and
splitting it on characters that are not Unicode letters/numbers or a dash.
We currently do not remove diacritics so these will need to be
converted to ASCII by the user.
If a text query involves only non-letter/number/dash characters,
the filter will not be well-defined and will be ignored when
constructing SQL statements.

For convenience, a non-empty character vector may be used in ``query``.
- A list of length 1 is treated as shorthand for a text
query with default arguments in :py:func:`~.define_text_query`.
- A list of length greater than 1 is treated as
shorthand for an `AND` operation on default text queries for
each of the individual strings.

See Also:
    :py:func:`~gypsum_client.fetch_metadata_database.fetch_metadata_database`,
    to download and cache the database files.

    See `metadata index <https://github.com/ArtifactDB/bioconductor-metadata-index>`_,
    for details on the SQLite file contents and table structure.
"""

import json
import re
import sqlite3
from typing import Dict, List, Optional, Union

__author__ = "chatGPT"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

#####
## I'm tired, so used chatGPT to test out its R to Python capabilities.
## It did hallucinate a bit but otherwise did a pretty good job!
#####


class GypsumSearchClause:
    def __init__(
        self,
        type: str,
        text: Optional[str] = None,
        field: Optional[str] = None,
        partial: bool = False,
        children: Optional[List] = None,
        child: Optional[object] = None,
    ):
        self.type = type
        self.text = text
        self.field = field
        self.partial = partial
        self.children = children if children is not None else []
        self.child = child

    def __and__(self, other):
        return GypsumSearchClause(type="and", children=[self, other])

    def __or__(self, other):
        return GypsumSearchClause(type="or", children=[self, other])

    def __invert__(self):
        return GypsumSearchClause(type="not", child=self)


def search_metadata_text(
    path: str,
    query: Union[str, List[str], GypsumSearchClause],
    latest: bool = True,
    include_metadata: bool = True,
) -> List[Dict]:
    """Text search on the metadata database.

    Perform a text search on a SQLite database containing
    metadata from the gypsum backend. This is based on a precomputed
    tokenization of all string properties in each metadata document;
    see `metadata index <https://github.com/ArtifactDB/bioconductor-metadata-index>`_
    for details.

    Examples:
        - Search all metadata for a keyword:

        .. code-block:: python

            search_metadata_text(
                sqlite_path,
                ["mikoto"],
                include_metadata=False,
                latest=False
            )

        - Search for metadata containing multiple keywords (`AND` operation):

        .. code-block:: python

            search_metadata_text(
                sqlite_path,
                ["sakugawa", "judgement"],
                include_metadata=False,
                latest=False
            )

            # or use the ``&`` operation
            query = define_text_query("sakugawa") & define_text_query("judgement")
            result = search_metadata_text(
                sqlite_path,
                query,
                include_metadata=False,
                latest=False
            )

        - Search for metadata container either of the keywords (`OR` operation):

        .. code-block:: python

            # use the ``|`` operation
            query = define_text_query("uiharu") | define_text_query("rank")
            result = search_metadata_text(
                sqlite_path,
                query,
                include_metadata=False,
                latest=False
            )

    Args:
        path:
            Path to the SQLite file, usually obtained
            by :py:func:`~gypsum_client.fetch_assets.fetch_metadata`.

        query:
            List of keywords specifying the query to execute.

            May be :py:class:`~.GypsumSearchClause` class generated by
            :py:func:`~.define_text_query`.

        latest:
            Whether to only search in the latest version for each
            asset. Defaults to True.

        include_metadata:
            Whether metadata should be returned.
            Defaults to True.

    Returns:
        Results matching the query.

    """
    where = search_metadata_text_filter(query)
    cond = where["where"]
    params = where["parameters"]

    conn = sqlite3.connect(path)
    conn.row_factory = sqlite3.Row
    try:
        stmt = "SELECT versions.project AS project, versions.asset AS asset, versions.version AS version, path"

        if include_metadata:
            stmt += ", json_extract(metadata, '$') AS metadata"

        if not latest:
            stmt += ", versions.latest AS latest"

        stmt += " FROM paths LEFT JOIN versions ON paths.vid = versions.vid"

        if latest:
            cond.append("versions.latest = 1")

        if cond:
            stmt += " WHERE " + " AND ".join(cond)
            cursor = conn.execute(stmt, params)
        else:
            cursor = conn.execute(stmt)

        results = [dict(row) for row in cursor.fetchall()]
        if include_metadata:
            for result in results:
                result["metadata"] = json.loads(result["metadata"])

        return results
    finally:
        conn.close()


def define_text_query(
    text: str, field: Optional[str] = None, partial: bool = False
) -> GypsumSearchClause:
    """Define a query.

    Args:
        text:
            Text to search by.

        field:
            Name of the metadata field to search by.

            If `None`, search is performed on all metadata fields.

        partial:
            Whether text contains SQLite wild cards for partial matches.

            Defaults to False.

    Returns:
        `GypsumSearchClause` defining the search.
    """
    return GypsumSearchClause(type="text", text=text, field=field, partial=partial)


def search_metadata_text_filter(
    query: Union[str, List[str], GypsumSearchClause], pid_name: str = "paths.pid"
) -> Dict[str, Union[str, List]]:
    query = sanitize_query(query)

    if query is None:
        return {"where": [], "parameters": {}}

    env = {"parameters": {}}
    cond = build_query(query, pid_name, env)
    return {"where": cond, "parameters": env["parameters"]}


def sanitize_query(
    query: Union[str, List[str], GypsumSearchClause],
) -> Optional[GypsumSearchClause]:
    if isinstance(query, list):
        if len(query) > 1:
            query = GypsumSearchClause(
                type="and", children=[define_text_query(q) for q in query]
            )
        elif len(query) == 1:
            query = define_text_query(query[0])
        else:
            raise ValueError("'query' must have atleast 1 element.")

    if isinstance(query, str):
        query = define_text_query(query)

    if query.type == "not":
        query.child = sanitize_query(query.child)
        if query.child is None:
            return None
        return query

    if query.type != "text":
        query.children = [
            sanitize_query(child)
            for child in query.children
            if sanitize_query(child) is not None
        ]

        if not query.children:
            return None

        if len(query.children) == 1:
            return query.children[0]

        return query

    extras = "%" if query.partial else ""
    text_tokens = re.split(
        r"[^a-zA-Z0-9" + re.escape(extras) + r"-]", query.text.lower()
    )
    text_tokens = [token for token in text_tokens if token]
    if not text_tokens:
        return None

    children = [
        define_text_query(token, field=query.field, partial=query.partial)
        for token in text_tokens
    ]

    if len(children) == 1:
        return children[0]

    return GypsumSearchClause(type="and", children=children)


def add_query_parameter(env: Dict, value: str) -> str:
    param_name = f"p{len(env['parameters'])}"
    env["parameters"][param_name] = value
    return f":{param_name}"


def build_query(query: GypsumSearchClause, name: str, env: Dict) -> List[str]:
    if query.type == "text":
        param_name = add_query_parameter(env, query.text)
        match_str = f"tokens.token {'LIKE' if query.partial else '='} {param_name}"
        if query.field:
            field_param = add_query_parameter(env, query.field)
            return [
                f"{name} IN (SELECT pid FROM links LEFT JOIN tokens ON tokens.tid = links.tid LEFT JOIN fields ON fields.fid = links.fid WHERE {match_str} AND fields.field = {field_param})"
            ]
        return [
            f"{name} IN (SELECT pid FROM links LEFT JOIN tokens ON tokens.tid = links.tid WHERE {match_str})"
        ]

    if query.type == "not":
        return [f"NOT ({build_query(query.child, name, env)[0]})"]

    if query.type == "and":
        return [
            " AND ".join(build_query(child, name, env)[0] for child in query.children)
        ]

    if query.type == "or":
        is_text = [child for child in query.children if child.type == "text"]
        non_text = [child for child in query.children if child.type != "text"]

        text_clauses = []
        for child in is_text:
            param_name = add_query_parameter(env, child.text)
            match_str = f"tokens.token {'LIKE' if child.partial else '='} {param_name}"
            if child.field:
                field_param = add_query_parameter(env, child.field)
                text_clauses.append(f"({match_str} AND fields.field = {field_param})")
            else:
                text_clauses.append(match_str)

        if text_clauses:
            text_query = " OR ".join(text_clauses)
            text_query = f"{name} IN (SELECT pid FROM links LEFT JOIN tokens ON tokens.tid = links.tid WHERE {text_query})"
        else:
            text_query = ""

        non_text_clauses = " OR ".join(
            build_query(child, name, env)[0] for child in non_text
        )

        if len(non_text) > 0:
            return [f"({text_query} OR {non_text_clauses})"]
        else:
            return [f"({text_query})"]

    raise ValueError(f"Unsupported query type: {query.type}")



from typing import List, Union
from urllib.parse import quote_plus

import requests

from ._utils import _remove_slash_url, _sanitize_uploaders
from .auth import access_token
from .config import REQUESTS_MOD
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def create_project(
    project: str,
    owners: Union[str, List[str]],
    uploaders: List[str] = [],
    baseline: int = None,
    growth_rate: int = None,
    year: int = None,
    url: str = rest_url(),
    token: str = None,
):
    """Create a new project with the associated permissions.

    See Also:
        :py:func:`~gypsum_client.remove_operations.remove_project`,
        to remove the project.

    Example:

        .. code-block:: python

            createProject(
                "test-Py-create",
                owners="jkanche",
                uploaders=[{"id": "ArtifactDB-bot"}]
            )

    Args:
        project:
            Project name.

        owners:
            List of GitHub users or organizations that are owners of this
            project.

            May also be a string containing the Github user or organization.

        uploaders:
            List of authorized uploaders for this project.
            Defaults to an empty list.

            Checkout :py:func:`~gypsum_client.fetch_assets.fetch_permissions`
            for more info.

        baseline:
            Baseline quote in bytes.

        growth_rate:
            Expected annual growth rate in bytes.

        year:
            Year of project creation.

        url:
            URL to the gypsum REST API.

        token:
            GitHub access token to authenticate with the gypsum REST API.
            The token must refer to a gypsum administrator account.
    """
    url = _remove_slash_url(url)
    uploaders = _sanitize_uploaders(uploaders) if uploaders is not None else []

    if token is None:
        token = access_token()

    quota = {}
    if baseline is not None:
        quota["baseline"] = baseline

    if growth_rate is not None:
        quota["growth_rate"] = growth_rate

    if year is not None:
        quota["year"] = year

    if isinstance(owners, str):
        owners = [owners]

    body = {"permissions": {"owners": owners, "uploaders": uploaders}}
    if len(quota) > 0:
        body["quota"] = quota

    req = requests.post(
        f"{url}/create/{quote_plus(project)}",
        json=body,
        headers={"Authorization": "Bearer " + token},
        verify=REQUESTS_MOD["verify"],
    )
    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to create a project, {req.status_code} and reason: {req.text}"
        ) from e



import atexit
import json
import os
import re
import shutil
from multiprocessing import Pool
from typing import Optional

from ._utils import (
    BUCKET_CACHE_NAME,
    _acquire_lock,
    _release_lock,
    _sanitize_path,
    _save_file,
)
from .cache_directory import cache_directory
from .config import REQUESTS_MOD
from .list_operations import list_files
from .resolve_links import resolve_links
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def _save_file_wrapper(args):
    x, project, asset, version, destination, overwrite, url, verify = args
    path = os.path.join(project, asset, version, x)
    dest = os.path.join(destination, x)
    _save_file(path=path, destination=dest, overwrite=overwrite, url=url, verify=verify)


def save_version(
    project: str,
    asset: str,
    version: str,
    cache_dir: Optional[str] = cache_directory(),
    overwrite: bool = False,
    relink: bool = True,
    concurrent: int = 1,
    url: str = rest_url(),
) -> str:
    """Download all files associated with a version of an asset
    of a project from the gypsum bucket.

    See Also:

        :py:func:`~.save_file`, to save a single file.

    Example:

        .. code-block:: python

            out <- save_version("test-R", "basic", "v1")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        cache_dir:
            Path to the cache directory.

        overwrite:
            Whether to overwrite existing file in cache.

        relink:
            Whether links should be resolved, see :py:func:`~resolve_links`.
            Defaults to True.

        concurrent:
            Number of concurrent downloads.
            Defaults to 1.

    Returns:
        Path to the local directory where the files are downloaded to.
    """

    _acquire_lock(cache_dir, project, asset, version)

    def release_lock_wrapper():
        _release_lock(project, asset, version)

    atexit.register(release_lock_wrapper)
    destination = os.path.join(cache_dir, BUCKET_CACHE_NAME, project, asset, version)

    # If this version's directory was previously cached in its complete form, we skip it.
    completed = os.path.join(cache_dir, "status", project, asset, version, "COMPLETE")
    if not os.path.exists(completed) or overwrite:
        listing = list_files(project, asset, version, url=url)

        if concurrent <= 1:
            for file in listing:
                _save_file_wrapper(
                    (
                        file,
                        project,
                        asset,
                        version,
                        destination,
                        overwrite,
                        url,
                        REQUESTS_MOD["verify"],
                    )
                )
        else:
            _args = [
                (
                    file,
                    project,
                    asset,
                    version,
                    destination,
                    overwrite,
                    url,
                    REQUESTS_MOD["verify"],
                )
                for file in listing
            ]
            with Pool(concurrent) as pool:
                pool.map(_save_file_wrapper, _args)

        if relink:
            resolve_links(
                project,
                asset,
                version,
                cache_dir=cache_dir,
                overwrite=overwrite,
                url=url,
            )

        # Marking it as complete.
        os.makedirs(os.path.dirname(completed), exist_ok=True)
        with open(completed, "w"):
            pass

    return destination


def _resolve_single_link(
    project: str,
    asset: str,
    version: str,
    path: str,
    cache: str,
    overwrite: bool,
    url: str,
) -> Optional[str]:
    if "/" in path:
        lpath = f"{os.path.dirname(path)}/..links"
    else:
        lpath = "..links"

    lobject = f"{project}/{asset}/{version}/{lpath}"
    ldestination = os.path.join(
        cache, BUCKET_CACHE_NAME, project, asset, version, lpath
    )

    _saved = _save_file(
        lobject, ldestination, overwrite=overwrite, url=url, error=False
    )

    if not _saved:
        return None

    with open(ldestination, "r") as f:
        link_info = json.load(f)

    base = re.sub(r".*/", "", path)

    if base not in link_info:
        return None

    target = link_info[base]
    if "ancestor" in target:
        target = target["ancestor"]

    tobject = (
        f"{target['project']}/{target['asset']}/{target['version']}/{target['path']}"
    )
    tdestination = os.path.join(
        cache,
        BUCKET_CACHE_NAME,
        target["project"],
        target["asset"],
        target["version"],
        target["path"],
    )

    _save_file(tobject, tdestination, overwrite=overwrite, url=url)
    return tdestination


def save_file(
    project: str,
    asset: str,
    version: str,
    path: str,
    cache_dir: Optional[str] = cache_directory(),
    overwrite: bool = False,
    url: str = rest_url(),
):
    """Save a file from a version of a project asset.

    Download a file from the gypsum bucket, for a version of
    an asset of a project.

    See Also:

        :py:func:`~.save_version`, to save all files associated
        with a version.

    Example:

        .. code-block:: python

            out <- save_version("test-R", "basic", "v1", "blah.txt")


    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        path:
            Suffix of the object key for the file of interest,
            i.e., the relative ``path`` inside the version's `
            `subdirectory``.

            The full object key is defined as
            ``{project}/{asset}/{version}/{path}``.

        cache_dir:
            Path to the cache directory.

        overwrite:
            Whether to overwrite existing file in cache.

        url:
            URL to the gypsum compatible API.

    Returns:
        The destintion file path where the file is downloaded to in the local
        file system.
    """

    _acquire_lock(cache_dir, project, asset, version)

    def release_lock_wrapper():
        _release_lock(project, asset, version)

    atexit.register(release_lock_wrapper)

    object_key = f"{project}/{asset}/{version}/{_sanitize_path(path)}"
    destination = os.path.join(
        cache_dir, BUCKET_CACHE_NAME, project, asset, version, path
    )

    found = _save_file(
        object_key, destination, overwrite=overwrite, url=url, error=False
    )

    if not found:
        link = _resolve_single_link(
            project, asset, version, path, cache_dir, overwrite=overwrite, url=url
        )

        if link is None:
            raise ValueError(f"'{path}' does not exist in the bucket.")

        try:
            os.link(link, destination)
        except Exception:
            try:
                os.symlink(link, destination)
            except Exception:
                try:
                    shutil.copy(link, destination)
                except Exception as e:
                    raise ValueError(
                        f"Failed to resolve link for '{path}': {e}."
                    ) from e

    return destination



import requests

from ._utils import _list_for_prefix
from .config import REQUESTS_MOD
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def list_projects(url: str = rest_url()) -> list:
    """List all projects in the gypsum backend.

    Example:

        .. code-block:: python

            all_prjs = list_projects()

    Args:
        url:
            URL to the gypsum compatible API.

    Returns:
        List of project names.
    """
    return _list_for_prefix(prefix=None, url=url)


def list_assets(project: str, url: str = rest_url()) -> list:
    """List all assets in a project.

    Example:

        .. code-block:: python

            all_assets = list_assets("test-R")

    Args:
        project:
            Project name.

        url:
            URL to the gypsum compatible API.

    Returns:
        List of asset names.
    """
    return _list_for_prefix(f"{project}/", url=url)


def list_versions(project: str, asset: str, url: str = rest_url()) -> list:
    """List all versions for a project asset.

    Example:

        .. code-block:: python

            all_vers = list_versions("test-R", "basic")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        url:
            URL to the gypsum compatible API.

    Returns:
        List of versions.
    """
    return _list_for_prefix(f"{project}/{asset}/", url=url)


def list_files(
    project: str,
    asset: str,
    version: str,
    prefix: str = None,
    include_dot: bool = True,
    url: str = rest_url(),
) -> list:
    """List all files for a specified version of a project and asset.

    Example:

        .. code-block:: python

            all_files = list_files("test-R", "basic", "v1")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        prefix:
            Prefix for the object key.

            If provided. a file is only listed if its object key starts with
            ``{project}/{asset}/{version}/{prefix}``.

            Defaults to None and all associated files with this version of the
            asset in the specified project are listed.

        include_dot:
            Whether to list files with ``..`` in their names.

        url:
            URL to the gypsum compatible API.

    Returns:
        List of relative paths of files associated with the versioned asset.
    """
    _prefix = f"{project}/{asset}/{version}/"
    _trunc = len(_prefix)
    if prefix is not None:
        _prefix = f"{_prefix}{prefix}"

    req = requests.get(
        f"{url}/list",
        params={"recursive": "true", "prefix": _prefix},
        verify=REQUESTS_MOD["verify"],
    )
    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to list files in a project, {req.status_code} and reason: {req.text}"
        ) from e
    resp = req.json()

    resp = [val[_trunc:] for val in resp]

    if prefix is not None:
        resp = [val for val in resp if val.startswith(prefix)]

    if include_dot is False:
        resp = [val for val in resp if not val.startswith("..")]

    return resp



from gypsum_client.list_operations import (
    list_assets,
    list_files,
    list_projects,
    list_versions,
)

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_list_projects():
    projects = list_projects()
    assert "test-R" in projects


def test_list_assets():
    assets = list_assets("test-R")
    assert "basic" in assets


def test_list_versions():
    versions = list_versions("test-R", "basic")
    assert "v1" in versions
    assert "v2" in versions
    assert "v3" in versions


def test_list_files():
    in_basic = list_files("test-R", "basic", "v1")
    assert sorted(in_basic) == sorted(
        ["..summary", "..manifest", "blah.txt", "foo/bar.txt"]
    )

    in_basic = list_files("test-R", "basic", "v2")
    assert sorted(in_basic) == sorted(
        ["..summary", "..manifest", "..links", "foo/..links"]
    )

    in_basic = list_files("test-R", "basic", "v1", prefix="foo/")
    assert sorted(in_basic) == sorted(["foo/bar.txt"])

    in_basic = list_files("test-R", "basic", "v1", prefix="..")
    assert sorted(in_basic) == sorted(["..summary", "..manifest"])



from urllib.parse import quote_plus

import requests

from ._utils import (
    _remove_slash_url,
)
from .auth import access_token
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def refresh_latest(
    project: str, asset: str, url: str = rest_url(), token: str = None
) -> str:
    """Refresh the latest version.

    Recompute the latest version of a project's asset.
    This is useful on rare occasions where multiple simultaneous
    uploads cause the latest version to be slightly out of sync.

    See Also:
        :py:func:`~gypsum_client.fetch_operations.fetch_latest`,
        to fetch the latest version without recomputing.

    Example:

        .. code-block:: python

            ver = refresh_latest("test-R", "basic")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.

    Return:
        A string containing the latest version.
    """

    if token is None:
        token = access_token()

    url = _remove_slash_url(url)
    _key = f"{quote_plus(project)}/{quote_plus(asset)}"
    req = requests.post(
        f"{url}/refresh/latest/{_key}",
        headers={"Authorization": f"Bearer {token}"},
    )

    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to refresh latest version, {req.status_code} and reason: {req.text}"
        ) from e

    return req.json()["version"]


def refresh_usage(project: str, url: str = rest_url(), token: str = None) -> int:
    """Refresh the usage quota of a project.

    Recompute the usage of a project.
    This is useful on rare occasions where multiple simultaneous
    uploads cause the usage calculations to be out of sync.

    See Also:
        :py:func:`~gypsum_client.fetch_operations.fetch_usage`,
        to fetch the usage without recomputing.

    Example:

        .. code-block:: python

            ver = refresh_usage("test-R", "basic")

    Args:
        project:
            Project name.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.

    Returns:
        The total quota usage of this project, in bytes.
    """

    if token is None:
        token = access_token()

    url = _remove_slash_url(url)
    _key = f"{quote_plus(project)}"
    req = requests.post(
        f"{url}/refresh/usage/{_key}",
        headers={"Authorization": f"Bearer {token}"},
    )

    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to refresh quota usage, {req.status_code} and reason: {req.text}"
        ) from e

    return req.json()["total"]



import json
import os
import re
import shutil
import tempfile
from datetime import datetime
from typing import Optional
from urllib.parse import quote_plus

import requests
from filelock import FileLock

from .config import REQUESTS_MOD

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def _remove_slash_url(url: str):
    if url.endswith("/"):
        url = url.rstrip("/")

    return url


def _list_for_prefix(
    prefix: str,
    url: str,
    recursive: bool = False,
    include_dot: bool = False,
    only_dirs: bool = True,
):
    url = url + "/list"

    qparams = {"recursive": "true" if recursive is True else "false"}
    if prefix is not None:
        qparams["prefix"] = prefix

    req = requests.get(url, params=qparams, verify=REQUESTS_MOD["verify"])
    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to access files from API, {req.status_code} and reason: {req.text}"
        ) from e

    resp = req.json()
    if only_dirs is True:
        resp = [val for val in resp if val.endswith("/")]

    if prefix is not None:
        resp = [val.replace(prefix, "") for val in resp if val.startswith(prefix)]

    if include_dot is False:
        resp = [_remove_slash_url(val) for val in resp if not val.startswith("..")]

    return resp


def _fetch_json(path: str, url: str):
    full_url = f"{url}/file/{quote_plus(path)}"

    req = requests.get(full_url, verify=REQUESTS_MOD["verify"])
    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to access json from API, {req.status_code} and reason: {req.text}"
        ) from e

    return req.json()


BUCKET_CACHE_NAME = "bucket"


def _fetch_cacheable_json(
    project: str,
    asset: str,
    version: str,
    path: str,
    cache: str,
    url: str,
    overwrite: bool,
):
    bucket_path = f"{project}/{asset}/{version}/{path}"

    if cache is None:
        return _fetch_json(bucket_path, url=url)
    else:
        _out_path = os.path.join(
            cache, BUCKET_CACHE_NAME, project, asset, version, path
        )

        _save_file(bucket_path, destination=_out_path, overwrite=overwrite, url=url)

        with open(_out_path, "r") as jf:
            return json.load(jf)


def _save_file(
    path: str,
    destination: str,
    overwrite: bool,
    url: str,
    error: bool = True,
    verify: Optional[bool] = None,
):
    if overwrite is True or not os.path.exists(destination):
        os.makedirs(os.path.dirname(destination), exist_ok=True)

        _lock = FileLock(destination + ".LOCK")
        with _lock:
            with tempfile.NamedTemporaryFile(
                dir=os.path.dirname(destination), delete=False
            ) as tmp_file:
                try:
                    full_url = f"{url}/file/{quote_plus(path)}"

                    if verify is None:
                        verify = REQUESTS_MOD["verify"]

                    req = requests.get(full_url, stream=True, verify=verify)
                    try:
                        req.raise_for_status()
                    except Exception as e:
                        raise Exception(
                            f"Failed to save file from API, {req.status_code} and reason: {req.text}"
                        ) from e

                    for chunk in req.iter_content(chunk_size=None):
                        tmp_file.write(chunk)
                except Exception as e:
                    if error:
                        raise Exception(f"Failed to save '{path}'; {str(e)}.") from e
                    else:
                        return False

                # Rename the temporary file to the destination
                shutil.move(tmp_file.name, destination)

    return True


def _cast_datetime(x):
    # Remove fractional seconds.
    if "." in x:
        x = x.split(".")[0]

    if x.endswith("Z"):
        x = x[:-1]

    return datetime.strptime(x, "%Y-%m-%dT%H:%M:%S")


def _rename_file(src: str, dest: str):
    try:
        os.rename(src, dest)
    except OSError:
        try:
            # If renaming fails, try copying
            shutil.copy(src, dest)
            os.remove(src)  # Remove the original file after copying
        except Exception as e:
            raise RuntimeError(
                f"Cannot move temporary file for '{src}' to its destination '{dest}': {e}."
            ) from e


def _download_and_rename_file(url: str, dest: str):
    tmp = tempfile.NamedTemporaryFile(dir=os.path.dirname(dest), delete=False).name
    req = requests.get(url, stream=True, verify=REQUESTS_MOD["verify"])

    with open(tmp, "wb") as f:
        for chunk in req.iter_content():
            f.write(chunk)

    _rename_file(tmp, dest)


IS_LOCKED = {"locks": {}}


def _acquire_lock(cache: str, project: str, asset: str, version: str):
    _key = f"{project}/{asset}/{version}"

    if _key in IS_LOCKED["locks"] and IS_LOCKED["locks"][_key] is None:
        _path = os.path.join(cache, "status", project, asset, version)
        os.makedirs(os.path.dirname(_path), exist_ok=True)

        _lock = FileLock(_path + ".LOCK")
        _lock.acquire()
        IS_LOCKED["locks"][_key] = _lock


def _release_lock(project: str, asset: str, version: str):
    _key = f"{project}/{asset}/{version}"

    if _key in IS_LOCKED["locks"] and IS_LOCKED["locks"][_key] is not None:
        _lock = IS_LOCKED["locks"][_key]
        _lock.release()
        del IS_LOCKED["locks"][_key]


def _sanitize_path(x):
    if os.name == "nt":
        x = re.sub(r"\\\\", "/", x)

    x = re.sub(r"//+", "/", x)
    return x


def _sanitize_uploaders(uploaders: list):
    for current in uploaders:
        if "until" in current:
            _cur_until = current["until"]
            if isinstance(_cur_until, str):
                _cur_until = _cast_datetime(_cur_until)

            current["until"] = _cur_until.isoformat().replace("+00:00", "Z")

    return uploaders


# from https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


# from https://stackoverflow.com/questions/6108330/checking-for-interactive-shell-in-a-python-script
def _is_interactive():
    import sys

    return sys.__stdin__.isatty() or is_notebook()



import os
from pathlib import Path
from typing import Optional

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

CURRENT_CACHE_DIRECTORY = None


def cache_directory(dir: Optional[str] = None):
    """Cache directory.

    Specify the cache directory in the local filesystem
    for gypsum-related data.

    If the ``GYPSUM_CACHE_DIR`` environment variable is set
    before the first call to ``cache_directory()``, it is used
    as the initial location of the cache directory.
    Otherwise, the initial location is set to user's home
    directory defined by ``Path.home()``.

    Args:
        dir:
            Path to the cache directory.
            If `None`, a default cache location is used.

    Returns:
        Path to the cache directory.
    """
    global CURRENT_CACHE_DIRECTORY

    if CURRENT_CACHE_DIRECTORY is None:
        _from_env = os.environ.get("GYPSUM_CACHE_DIR", None)
        if _from_env is not None:
            if not os.path.exists(_from_env):
                raise FileNotFoundError(
                    f"Path {_from_env} does not exist or is not accessible."
                )

            CURRENT_CACHE_DIRECTORY = _from_env
        else:
            CURRENT_CACHE_DIRECTORY = os.path.join(str(Path.home()), "gypsum", "cache")
            os.makedirs(CURRENT_CACHE_DIRECTORY, exist_ok=True)

    if dir is not None:
        if not os.path.exists(dir):
            raise FileNotFoundError(f"Path {dir} does not exist or is not accessible.")
        CURRENT_CACHE_DIRECTORY = dir

    return CURRENT_CACHE_DIRECTORY



from urllib.parse import quote_plus

import requests

from ._utils import _remove_slash_url, _sanitize_uploaders
from .auth import access_token
from .fetch_operations import fetch_permissions
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def set_quota(
    project: str,
    baseline: int = None,
    growth_rate: int = None,
    year: int = None,
    url: str = rest_url(),
    token: str = None,
):
    """
    Set the quota for a project.

    See Also:
        :py:func:`~gypsum_client.fetch_operations.fetch_quota`,
        to fetch the usage details.

    Example:

        .. code-block:: python

            create_project("test-Py-quota", owners=["jkanche"])

            set_quota(
                "test-Py-quota",
                baseline=1234,
                growth_rate=5678,
                year=2020
            )


    Args:
        project:
            The project name.

        baseline:
            Baseline quote, in bytes.
            If `None`, no change is made.

        growth_rate:
            Expected annual growth rate, in bytes.
            If `None`, no change is made.

        year:
            Year of project creation.
            If `None`, no change is made.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.
    """

    if token is None:
        token = access_token()

    url = _remove_slash_url(url)

    body = {}
    if baseline is not None:
        body["baseline"] = baseline

    if growth_rate is not None:
        body["growth_rate"] = growth_rate

    if year is not None:
        body["year"] = year

    endpoint = f"{url}/quota/{quote_plus(project)}"
    headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}

    req = requests.put(endpoint, json=body, headers=headers)

    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to set quota, {req.status_code} and reason: {req.text}."
        ) from e


def set_permissions(
    project: str,
    owners: str = None,
    uploaders: str = None,
    append: bool = True,
    url: str = rest_url(),
    token: str = None,
):
    """
    Set the owner and uploader permissions for a project.

    See Also:
        :py:func:`~gypsum_client.fetch_operations.fetch_permission`,
        to fetch the permissions for a project.

    Example:

        .. code-block:: python

            create_project("test-Py-perms", owners=["jkanche"])

            until = (
                (datetime.now() + timedelta(seconds=1000000))
                .replace(microsecond=0)
            )

            set_permissions(
                "test-Py-perms",
                owners=["LTLA"],
                uploaders=[{"id": "LTLA", "until": until}]
            )

    Args:
        project:
            The project name.

        owners:
            List of GitHub users or organizations that are owners of this
            project.

            If `None`, no change is made.

        uploaders:
            List of authorized uploaders for this project.

            If `None`, no change is made.

        append:
            Whether to append owners and uploaders to the existing ones.

            If `False`, the provided owners and uploaders replace the existing values.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.
    """
    if token is None:
        token = access_token()

    url = _remove_slash_url(url)

    perms = {}
    if append is True:
        old_perms = fetch_permissions(project, url=url)

        if owners is not None:
            perms["owners"] = list(set(old_perms["owners"]).union(owners))

        if uploaders is not None:
            perms["uploaders"] = old_perms["uploaders"] + uploaders
    else:
        if owners is not None:
            perms["owners"] = owners

        if uploaders is not None:
            perms["uploaders"] = uploaders

    if "uploaders" in perms:
        perms["uploaders"] = _sanitize_uploaders(perms["uploaders"])

    endpoint = f"{url}/permissions/{quote_plus(project)}"
    headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}

    req = requests.put(endpoint, json=perms, headers=headers)

    try:
        req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to set permissions, {req.status_code} and reason: {req.text}."
        ) from e



import atexit
import os
import shutil
from typing import Optional

from ._utils import (
    BUCKET_CACHE_NAME,
    _acquire_lock,
    _release_lock,
)
from .cache_directory import cache_directory
from .fetch_operations import fetch_manifest
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def resolve_links(
    project: str,
    asset: str,
    version: str,
    cache_dir: Optional[str] = cache_directory(),
    overwrite: str = False,
    url: str = rest_url(),
):
    """Resolve links in the cache directory.

    Create hard links (or copies, if filesystem links
    are not supported) for linked-from files to their
    link destinations.

    Example:

        .. code-block:: python

            cache = tempfile()

            save_version("test-R", "basic", "v3", relink=False, cache_dir=cache)
            list_files(cache_dir, recursive=True, all_files=True)

            resolve_links("test-R", "basic", "v3", cache_dir=cache)

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        cache_dir:
            Path to the cache directory.

        overwrite:
            Whether to overwrite existing file in cache.

        url:
            URL to the gypsum compatible API.

    Returns:
        True if all links are resolved.
    """
    from .save_operations import save_file

    _acquire_lock(cache_dir, project, asset, version)

    def release_lock_wrapper():
        _release_lock(project, asset, version)

    atexit.register(release_lock_wrapper)

    # destination = os.path.join(cache_dir, BUCKET_CACHE_NAME, project, asset, version)
    manifests = {}

    self_manifest = fetch_manifest(
        project, asset, version, cache_dir=cache_dir, url=url
    )
    manifests["/".join([project, asset, version])] = self_manifest

    for kmf in self_manifest.keys():
        entry = self_manifest[kmf]
        if entry.get("link") is None:
            continue

        old_loc = os.path.join(project, asset, version, kmf)
        if os.path.exists(old_loc) and not overwrite:
            continue

        link_data = entry["link"]
        if link_data.get("ancestor") is not None:
            link_data = link_data["ancestor"]

        out = save_file(
            link_data["project"],
            link_data["asset"],
            link_data["version"],
            link_data["path"],
            cache_dir=cache_dir,
            url=url,
            overwrite=overwrite,
        )
        old_path = os.path.join(cache_dir, BUCKET_CACHE_NAME, old_loc)

        try:
            os.unlink(old_path)
        except Exception:
            pass

        os.makedirs(os.path.dirname(old_path), exist_ok=True)

        try:
            os.link(out, old_path)
        except Exception:
            try:
                os.symlink(out, old_path)
            except Exception:
                try:
                    shutil.copy(out, old_path)
                except Exception as e:
                    raise ValueError(f"Failed to resolve link for '{kmf}': {e}") from e

    return True



import json
from typing import Optional, Union

from jsonschema import validate as json_validate

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def validate_metadata(
    metadata: Union[str, dict], schema: str, stringify: Optional[bool] = None
) -> bool:
    """Validate metadata against a JSON schema for a SQLite database.

    See Also:
        :py:func:`~gypsum_client.fetch_metadata_schema.fetch_metadata_schema`, to get
        the JSON schema.

        :py:func:`~gypsum_client.fetch_metadata_database.fetch_metadata_database`,
        to obtain the SQLite database of metadata.

    Example:

        .. code-block:: python

            _cache_dir = tempfile.mkdtemp()

            metadata = {
                "title": "Fatherhood",
                "description": "Luke ich bin dein Vater.",
                "sources": [{"provider": "GEO", "id": "GSE12345"}],
                "taxonomy_id": ["9606"],
                "genome": ["GRCm38"],
                "maintainer_name": "Darth Vader",
                "maintainer_email": "vader@empire.gov",
                "bioconductor_version": "3.10",
            }

            schema = fetch_metadata_schema(cache_dir=_cache_dir)
            validate_metadata(metadata, schema)

    Args:
        metadata:
            Metadata to be checked.

            Usually a dictionary, but may also be a JSON-formatted string.

        schema:
            Path to a schema.

        stringify:
            Whether to convert ``metadata`` to a JSON-formatted string.
            Defaults to True if ``metadata`` is not already a string.

    Returns:
        True if metadata conforms to schema.
    """
    if stringify is None:
        stringify = not isinstance(metadata, str)

    if stringify:
        metadata = json.dumps(metadata)

    with open(schema) as f:
        schema_data = json.load(f)

    try:
        json_validate(instance=json.loads(metadata), schema=schema_data)
    except Exception as e:
        raise ValueError(f"Metadata validation failed: {e}") from e

    return True



"""Fetch the metadata database.

This function will automatically check for updates to the SQLite files
and will download new versions accordingly. New checks are performed when one hour
or more has elapsed since the last check. If the check fails, a warning is raised
and the function returns the currently cached file.
"""

import os
import tempfile
import time
import warnings

import requests
from filelock import FileLock

from ._utils import _download_and_rename_file
from .cache_directory import cache_directory
from .config import REQUESTS_MOD

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

LAST_CHECK = {"req_time": None, "mod_time": None}


def fetch_metadata_database(
    name: str = "bioconductor.sqlite3",
    cache_dir: str = cache_directory(),
    overwrite: bool = False,
) -> str:
    """Fetch the SQLite database containing metadata from the gypsum backend.

    See `metadata index <https://github.com/ArtifactDB/bioconductor-metadata-index>`_
    for more details.

    Each database is generated by aggregating metadata across multiple assets
    and/or projects, and can be used to perform searches for interesting objects.

    See Also:
        :py:func:`~gypsum_client.fetch_metadata_schema.fetch_metadata_schema`, to get
        the JSON schema used to define the database tables.

    Example:

        .. code-block:: python

            sql_path = fetch_metadata_database()

    Args:
        name:
            Name of the database.
            This can be the name of any SQLite file published
            `here <https://github.com/ArtifactDB/bioconductor-metadata-index/releases/tag/latest>`_.

            Defaults to "bioconductor.sqlite3".

        cache_dir:
            Path to the cache directory.

            Defaults to None.

        overwrite:
            Whether to overwrite existing file.

            Defaults to False.

    Returns:
        Path to the downloaded database.
    """
    base_url = "https://github.com/ArtifactDB/bioconductor-metadata-index/releases/download/latest/"

    if cache_dir is None:
        cache_path = tempfile.NamedTemporaryFile(suffix=".sqlite3").name
    else:
        cache_dir = os.path.join(cache_dir, "databases")

        cache_path = os.path.join(cache_dir, name)
        if not os.path.exists(cache_path):
            os.makedirs(os.path.dirname(cache_path), exist_ok=True)

        if os.path.exists(cache_path) and not overwrite:
            old_lastmod_raw = None

            _lock = FileLock(cache_path + ".LOCK")
            if not _lock.is_locked:
                old_lastmod_raw = open(cache_path + ".modified").readlines()

            old_lastmod = float(old_lastmod_raw[0])
            new_lastmod = get_last_modified_date(base_url)

            if new_lastmod is not None and old_lastmod == new_lastmod:
                return cache_path

    _lock = FileLock(cache_path + ".LOCK")
    with _lock:
        mod_path = cache_path + ".modified"
        _download_and_rename_file(base_url + "modified", mod_path)
        _download_and_rename_file(base_url + name, cache_path)

    LAST_CHECK["req_time"] = get_current_unix_time()
    LAST_CHECK["mod_time"] = float(open(mod_path).readline())

    return cache_path


def get_current_unix_time():
    return time.time() * 1000  # milliseconds


def get_last_modified_date(base_url):
    curtime = get_current_unix_time()
    if (
        LAST_CHECK["req_time"] is not None
        and LAST_CHECK["req_time"] + 60 * 60 * 1000 >= curtime
    ):
        return LAST_CHECK["mod_time"]

    mod_time = None
    try:
        url = base_url + "modified"
        response = requests.get(url, verify=REQUESTS_MOD["verify"])
        mod_time = float(response.text)
    except Exception as e:
        warnings.warn(
            f"Failed to check the last modified timestamp: {str(e)}", UserWarning
        )

    if mod_time is not None:
        LAST_CHECK["req_time"] = curtime
        LAST_CHECK["mod_time"] = mod_time

    return mod_time



import os
import tempfile

import pytest
from gypsum_client.fetch_metadata_database import fetch_metadata_database, LAST_CHECK


__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def test_fetch_metadata_database():
    _cache_dir = tempfile.mkdtemp()

    path = fetch_metadata_database(cache_dir=_cache_dir)
    assert os.path.getsize(path) > 0
    assert isinstance(LAST_CHECK["req_time"], float)
    assert not isinstance(LAST_CHECK["req_time"], bool)
    assert not isinstance(LAST_CHECK["mod_time"], bool)
    assert isinstance(LAST_CHECK["mod_time"], float)

    # Uses the cache.
    with open(path, "w") as f:
        f.write("FOO_BAR")

    path2 = fetch_metadata_database(cache_dir=_cache_dir)
    assert path == path2

    assert open(path).read().strip() == "FOO_BAR"



import json
import sqlite3
import tempfile

from gypsum_client import define_text_query, search_metadata_text

__author__ = "Jayaram Kancherla, chatGPT"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

sqlite_path = tempfile.mkdtemp() + "/test.sqlite"


def setup_database():
    conn = sqlite3.connect(sqlite_path)
    cur = conn.cursor()

    cur.execute(
        """CREATE TABLE versions (
        vid INTEGER PRIMARY KEY,
        project TEXT,
        asset TEXT,
        version TEXT,
        latest BOOLEAN
    )"""
    )

    cur.execute(
        """CREATE TABLE paths (
        pid INTEGER PRIMARY KEY,
        vid INTEGER,
        path TEXT,
        metadata TEXT
    )"""
    )

    cur.execute(
        """CREATE TABLE tokens (
        tid INTEGER PRIMARY KEY,
        token TEXT
    )"""
    )

    cur.execute(
        """CREATE TABLE fields (
        fid INTEGER PRIMARY KEY,
        field TEXT
    )"""
    )

    cur.execute(
        """CREATE TABLE links (
        pid INTEGER,
        fid INTEGER,
        tid INTEGER
    )"""
    )

    versions = [
        (1, "foo", "bar", "1", False),
        (2, "foo", "bar", "2", False),
        (3, "foo", "bar", "3", True),
    ]
    cur.executemany("INSERT INTO versions VALUES (?, ?, ?, ?, ?)", versions)

    metadata = [
        {
            "first_name": "mikoto",
            "last_name": "misaka",
            "school": "tokiwadai",
            "ability": "railgun",
            "gender": "female",
            "comment": "rank 3",
        },
        {
            "first_name": "mitsuko",
            "last_name": "kongou",
            "school": "tokiwadai",
            "ability": "aerohand",
            "gender": "female",
        },
        {
            "first_name": "kuroko",
            "last_name": "shirai",
            "school": "tokiwadai",
            "ability": "teleport",
            "gender": "female",
            "affiliation": "judgement",
        },
        {
            "first_name": "misaki",
            "last_name": "shokuhou",
            "school": "tokiwadai",
            "ability": "mental out",
            "gender": "female",
            "comment": "rank 5",
        },
        {
            "first_name": "ruiko",
            "last_name": "saten",
            "school": "sakugawa",
            "gender": "female",
        },
        {
            "first_name": "kazari",
            "last_name": "uiharu",
            "school": "sakugawa",
            "gender": "female",
            "affiliation": "judgement",
        },
        {
            "first_name": "accelerator",
            "ability": "vector manipulation",
            "gender": "male",
            "comment": "rank 1",
        },
    ]

    paths = [
        (
            i + 1,
            (i % 3) + 1,
            f'{metadata[i]["first_name"]}.txt',
            json.dumps(metadata[i]),
        )
        for i in range(len(metadata))
    ]
    cur.executemany("INSERT INTO paths VALUES (?, ?, ?, ?)", paths)

    all_tokens = list(
        set(
            token
            for item in metadata
            for token in item.values()
            if isinstance(token, str)
            for token in token.split()
        )
    )
    cur.executemany(
        "INSERT INTO tokens (token) VALUES (?)", [(token,) for token in all_tokens]
    )

    all_fields = list(set(key for item in metadata for key in item.keys()))
    cur.executemany(
        "INSERT INTO fields (field) VALUES (?)", [(field,) for field in all_fields]
    )

    links = []
    for i, item in enumerate(metadata):
        for field, value in item.items():
            if isinstance(value, str):
                tokens = set(value.split())
                pid = i + 1
                fid = all_fields.index(field) + 1
                for token in tokens:
                    tid = all_tokens.index(token) + 1
                    links.append((pid, fid, tid))
    cur.executemany("INSERT INTO links VALUES (?, ?, ?)", links)

    conn.commit()
    conn.close()


setup_database()


def test_search_metadata_text_text_searches():
    result = search_metadata_text(
        sqlite_path, ["mikoto"], include_metadata=False, latest=False
    )

    assert len(result) == 1
    assert result[0]["path"] == "mikoto.txt"

    result1 = search_metadata_text(
        sqlite_path, "mikoto", include_metadata=False, latest=False
    )
    assert len(result1) == 1
    assert result1[0]["path"] == "mikoto.txt"

    result = search_metadata_text(
        sqlite_path, ["kuroko"], include_metadata=False, latest=False
    )
    assert len(result) == 1
    assert result[0]["path"] == "kuroko.txt"

    result = search_metadata_text(
        sqlite_path, ["TOKIWADAI"], include_metadata=False, latest=False
    )
    assert len(result) == 4

    paths = [r["path"] for r in result]
    assert sorted(paths) == sorted(
        ["mikoto.txt", "mitsuko.txt", "kuroko.txt", "misaki.txt"]
    )


def test_search_metadata_text_and_searches():
    result = search_metadata_text(
        sqlite_path, ["sakugawa", "judgement"], include_metadata=False, latest=False
    )
    assert len(result) == 1
    assert result[0]["path"] == "kazari.txt"

    query = define_text_query("sakugawa") & define_text_query("judgement")
    result = search_metadata_text(
        sqlite_path, query, include_metadata=False, latest=False
    )
    assert len(result) == 1
    assert result[0]["path"] == "kazari.txt"


def test_search_metadata_text_or_searches():
    query = define_text_query("uiharu") | define_text_query("rank")
    result = search_metadata_text(
        sqlite_path, query, include_metadata=False, latest=False
    )
    assert len(result) == 4

    paths = [r["path"] for r in result]
    assert sorted(paths) == sorted(
        ["mikoto.txt", "kazari.txt", "accelerator.txt", "misaki.txt"]
    )


def test_search_metadata_text_not_searches():
    query = ~define_text_query("uiharu")
    result = search_metadata_text(
        sqlite_path, query, include_metadata=False, latest=False
    )
    assert len(result) == 6

    paths = [r["path"] for r in result]
    assert sorted(paths) == sorted(
        [
            "mikoto.txt",
            "accelerator.txt",
            "misaki.txt",
            "mitsuko.txt",
            "kuroko.txt",
            "ruiko.txt",
        ]
    )


def test_search_metadata_text_combined_and_or_searches():
    result = search_metadata_text(
        sqlite_path, ["judgement", "sakugawa"], include_metadata=False, latest=False
    )
    assert len(result) == 1
    assert result[0]["path"] == "kazari.txt"


def test_search_metadata_text_respects_output_options():
    result = search_metadata_text(
        sqlite_path, ["female"], include_metadata=True, latest=False
    )
    assert len(result) == 6



import threading
import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer

import requests

from .config import REQUESTS_MOD

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

AUTH_CODE = None


def github_access_token(client_id, client_secret, authorization_url, token_url):
    """Get access token from GitHub.

    Users should look at :pu:func:`~.auth.access_token` function.

    This process temporarily spins up a server similar to capture the redirction url and
    extract the access token similar to how httr's `oauth_flow_auth_code` method.
    This allows us to programmatically capture access tokens from the url without the user
    having to manually input the code.

    Follows authentication flow described here -
    https://docs.github.com/en/apps/creating-github-apps/authenticating-with-a-github-app/generating-a-user-access-token-for-a-github-app.
    """

    class AuthorizationHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            global AUTH_CODE
            if self.path.startswith("/?code"):
                AUTH_CODE = self.path.split("/?code=")[1]

                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(
                    b"<h2>Authorization Code Received. Close this tab and go back to the python session.</h2>"
                )

                raise SystemExit

    def start_temp_server():
        server = HTTPServer(("localhost", 1410), AuthorizationHandler)
        server.serve_forever()

    _auth_url = (
        f"{authorization_url}?client_id={client_id}&redirect_url=http://localhost:1410"
    )

    server_thread = threading.Thread(target=start_temp_server)
    server_thread.start()
    webbrowser.open(_auth_url)
    server_thread.join()

    headers = {"accept": "application/json"}
    parameters = {
        "client_id": client_id,
        "client_secret": client_secret,
        "code": AUTH_CODE,
    }
    token_req = requests.post(
        token_url, headers=headers, json=parameters, verify=REQUESTS_MOD["verify"]
    )

    try:
        token_req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to access token, {token_req.status_code} and reason: {token_req.text}"
        ) from e

    token = token_req.json()["access_token"]

    return token



import os
from multiprocessing import Pool

import requests

from ._utils import _remove_slash_url
from .auth import access_token
from .cache_directory import cache_directory
from .config import REQUESTS_MOD
from .prepare_directory_for_upload import prepare_directory_upload
from .rest_url import rest_url
from .upload_api_operations import abort_upload, complete_upload, start_upload

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def upload_directory(
    directory: str,
    project: str,
    asset: str,
    version: str,
    cache_dir: str = cache_directory(),
    deduplicate: bool = True,
    probation: bool = False,
    url: str = rest_url(),
    token: str = None,
    concurrent: int = 1,
    abort_failed: bool = True,
) -> bool:
    """Upload a directory to the gypsum backend.

    This function is a wrapper around
    :py:func:`~gypsum_client.prepare_directory_for_upload.prepare_directory_upload`
    and :py:func:`~gypsum_client.upload_api_operations.start_upload` and others.

    The aim is to streamline the upload of a directory's contents
    when no customization of the file listing is required.

    Convenience method to upload a directory to the gypsum backend
    as a versioned asset of a project. This requires uploader permissions
    to the relevant project.

    Example:

        .. code-block:: python

            tmp_dir = tempfile.mkdtemp()

            with open(os.path.join(tmp, "blah.txt"), "w") as f:
                f.write("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

            os.makedirs(os.path.join(tmp, "foo"))
            with open(os.path.join(tmp, "foo", "bar.txt"), "w") as f:
                f.write("\n".join(map(str, range(1, 11))))

            upload_directory(
                tmp, "test-Py",
                "upload-dir", version="1"
            )

    Args:
        directory:
            Path to a directory containing the ``files`` to be uploaded.
            This directory is assumed to correspond to a version of an asset.

        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        cache_dir:
            Path to the cache for saving files, e.g., in
            :py:func:`~gypsum_client.save_operations.save_version`.

            Used to convert symbolic links to upload links,see
            :py:func:`~gypsum_client.prepare_directory_for_upload.prepare_directory_upload`.

        deduplicate:
            Whether the backend should attempt deduplication of ``files``
            in the immediately previous version.
            Defaults to True.

        probation:
            Whether to perform a probational upload.
            Defaults to False.

        url:
            URL of the gypsum REST API.

        token:
            GitHub access token to authenticate to the gypsum REST API.

        concurrent:
            Number of concurrent downloads.
            Defaults to 1.

        abort_failed:
            Whether to abort the upload on any failure.

            Setting this to `False` can be helpful for diagnosing upload problems.

    Returns:
        `True` if successfull, otherwise `False`.
    """

    if token is None:
        token = access_token()

    listing = prepare_directory_upload(directory, links="always", cache_dir=cache_dir)

    blob = start_upload(
        project=project,
        asset=asset,
        version=version,
        files=listing["files"],
        links=listing["links"],
        directory=directory,
        probation=probation,
        url=url,
        token=token,
    )

    success = False
    try:
        upload_files(blob, directory=directory, url=url, concurrent=concurrent)
        complete_upload(blob, url=url)
        success = True
    finally:
        if abort_failed and not success:
            abort_upload(blob)

    return success


def _upload_file_wrapper(args):
    file_info, directory, url, session_token = args
    _upload_file(file_info, directory, url, session_token)


def upload_files(
    init: dict, directory: str = None, url: str = rest_url(), concurrent: int = 1
):
    """Upload files in an initialized upload session for a version of an asset.

    Args:
        init:
            Dictionary containing ``file_urls`` and ``session_token``.
            This is typically the return value from
            :py:func:`~gypsum_client.start_upload.start_upload``.

        directory:
            Path to the directory containing files.
            Defaults to None, if files are part of the current working directory.

        url:
            URL of the gypsum REST API.

        concurrent:
            Number of concurrent uploads.
            Defaults to 1.
    """
    url = _remove_slash_url(url)

    if concurrent <= 1:
        for file_info in init["file_urls"]:
            _upload_file(file_info, directory, url, init["session_token"])
    else:
        with Pool(concurrent) as pool:
            _args = [
                (file_info, directory, url, init["session_token"])
                for file_info in init["file_urls"]
            ]
            pool.map(_upload_file_wrapper, _args)


def _upload_file(info: dict, directory: str, url: str, token: str):
    _path = info["path"]

    if directory is not None:
        _path = os.path.join(directory, _path)

    if info["method"] == "presigned":
        req_url = f"{url}{info['url']}"
        headers = {"Authorization": f"Bearer {token}"}
        res = requests.post(req_url, headers=headers, verify=REQUESTS_MOD["verify"])
        try:
            res.raise_for_status()
        except Exception as e:
            raise Exception(
                f"Failed to fetch pre-signed url, {res.status_code} and reason: {res.text}"
            ) from e

        presigned = res.json()

        req2_url = presigned["url"]
        headers2 = {"Content-MD5": presigned["md5sum_base64"]}
        with open(_path, "rb") as f:
            res2 = requests.put(
                req2_url, headers=headers2, data=f, verify=REQUESTS_MOD["verify"]
            )
            try:
                res2.raise_for_status()
            except Exception as e:
                raise Exception(
                    f"Failed to upload assets in the project, {res2.status_code} and reason: {res2.text}"
                ) from e
    else:
        raise ValueError(
            f"unknown upload method '{info['method']}' for file '{info['path']}'"
        )



import os

import pytest
from gypsum_client import (
    complete_upload,
    create_project,
    fetch_latest,
    fetch_manifest,
    fetch_usage,
    remove_asset,
    remove_project,
    remove_version,
    start_upload,
)

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

app_url = "https://gypsum-test.artifactdb.com"


@pytest.mark.skipif(
    "gh_token" not in os.environ, reason="GitHub token not in environment"
)
def test_remove_functions():
    gh_token = os.environ.get("gh_token", None)
    if gh_token is None:
        raise ValueError("GitHub token not in environment")

    remove_project("test-Py-remove", url=app_url, token=gh_token)

    create_project("test-Py-remove", owners=["jkanche"], url=app_url, token=gh_token)
    for v in ["v1", "v2"]:
        init = start_upload(
            project="test-Py-remove",
            asset="sacrifice",
            version=v,
            files=[],
            token=gh_token,
            url=app_url,
        )

        complete_upload(init, url=app_url)

    fetch_manifest("test-Py-remove", "sacrifice", "v2", url=app_url, cache_dir=None)
    remove_version("test-Py-remove", "sacrifice", "v2", url=app_url, token=gh_token)
    with pytest.raises(Exception):
        fetch_manifest("test-Py-remove", "sacrifice", "v2", url=app_url, cache_dir=None)

    assert fetch_latest("test-Py-remove", "sacrifice", url=app_url) == "v1"
    remove_asset("test-Py-remove", "sacrifice", url=app_url, token=gh_token)
    with pytest.raises(Exception):
        fetch_latest("test-Py-remove", "sacrifice", url=app_url)

    fetch_usage("test-Py-remove", url=app_url)
    remove_project("test-Py-remove", url=app_url, token=gh_token)
    with pytest.raises(Exception):
        fetch_usage("test-Py-remove", url=app_url)



from typing import Optional

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

CURRENT_REST_URL = "https://gypsum.artifactdb.com"


def rest_url(url: Optional[str] = None):
    """URL for the gypsum REST API.

    Get or set the URL for the gypsum REST API.

    Args:
        url:
            URL to the gypsum REST API.
            Defaults to None.

    Returns:
        String containing the URL to the gypsum REST API.
    """
    global CURRENT_REST_URL
    if url is not None:
        CURRENT_REST_URL = url

    return CURRENT_REST_URL



import os

from ._utils import (
    BUCKET_CACHE_NAME,
    _cast_datetime,
    _fetch_cacheable_json,
    _fetch_json,
)
from .cache_directory import cache_directory
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"


def fetch_latest(project: str, asset: str, url: str = rest_url()) -> str:
    """Fetch the latest version of a project's asset.

    See Also:
        :py:func:`~gypsum_client.refresh_operations.refresh_latest`,
        to refresh the latest version.

    Example:

        .. code-block:: python

            ver = fetch_latest("test-R", "basic")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        url:
            URL to the gypsum compatible API.

    Returns:
        Latest version of the project.
    """
    resp = _fetch_json(f"{project}/{asset}/..latest", url=url)
    return resp["version"]


def fetch_manifest(
    project: str,
    asset: str,
    version: str,
    cache_dir: str = cache_directory(),
    overwrite: bool = False,
    url: str = rest_url(),
) -> dict:
    """Fetch the manifest for a version of an asset of a project.

    Example:

        .. code-block:: python

            manifest = fetch_manifest("test-R", "basic", "v1")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        cache_dir:
            Path to the cache directory.

        overwrite:
            Whether to overwrite existing file in cache.

        url:
            URL to the gypsum compatible API.

    Returns:
        Dictionary containing the manifest for this version.
        Each element is named after the relative path of a file in this version.
        The value of each element is another list with the following fields:
        - ``size``, an integer specifying the size of the file in bytes.
        - ``md5sum``, a string containing the hex-encoded MD5 checksum of the file.
        - Optional ``link``,  a list specifying the link destination for a file.

        This contains the strings ``project``, ``asset``, ``version`` and ``path``.
        If the link destination is itself a link, an ``ancestor`` list will be
        present that specifies the final location of the file after resolving all intermediate links.
    """
    return _fetch_cacheable_json(
        project,
        asset,
        version,
        "..manifest",
        url=url,
        cache=cache_dir,
        overwrite=overwrite,
    )


def fetch_permissions(project: str, url: str = rest_url()) -> dict:
    """Fetch the permissions for a project.

    See Also:
        :py:func:`~gypsum_client.set_operations.set_permissions`,
        to update or modify the permissions.

    Example:

        .. code-block:: python

            perms = fetch_permissions("test-R")

    Args:
        project:
            Project name.

        url:
            URL to the gypsum compatible API.

    Returns:
        Dictionary containing the permissions for this project:
        - ``owners``, a character vector containing the GitHub users or
        organizations that are owners of this project.
        - ``uploaders``, a list of lists specifying the users or organizations
        who are authorzied to upload to this project.

        Each entry is a list with the following fields:
        - ``id``, a string containing the GitHub user or organization
        that is authorized to upload.
        - Optional ``asset``, a string containing the name of the asset
        that the uploader is allowed to upload to. If not provided, there is no
        restriction on the uploaded asset name.
        - Optional ``version``, a string containing the name of the version
        that the uploader is allowed to upload to.If not provided, there is
        no restriction on the uploaded version name.
        - Optional ``until``a POSIXct object containing the expiry date of this
        authorization. If not provided, the authorization does not expire.
        - Optional ``trusted``, whether the uploader is trusted.
        If not provided, defaults to False.

    """
    perms = _fetch_json(f"{project}/..permissions", url=url)

    for i, val in enumerate(perms["uploaders"]):
        if "until" in val:
            perms["uploaders"][i]["until"] = _cast_datetime(val["until"])

    return perms


def fetch_quota(project: str, url: str = rest_url()) -> dict:
    """Fetch the quota details for a project.

    See Also:
        :py:func:`~gypsum_client.set_operations.set_quota`,
        to update or modify the quota.

    Example:

        .. code-block:: python

            quota = fetch_quota("test-R")

    Args:
        project:
            Project name.

        url:
            URL to the gypsum compatible API.

    Returns:
        Dictionary containing ``baseline``, the baseline quota at time zero in bytes;
        ``growth_rate``, the annual growth rate for the quota in bytes;
        ``year``, the creation year (i.e., time zero) for this project.
    """
    return _fetch_json(f"{project}/..quota", url=url)


def fetch_summary(
    project: str,
    asset: str,
    version: str,
    cache_dir: str = cache_directory(),
    overwrite: bool = False,
    url: str = rest_url(),
) -> dict:
    """Fetch the summary for a version of an asset of a project.

    Example:

        .. code-block:: python

            summa = fetch_summary("test-R", "basic", "v1")

    Args:
        project:
            Project name.

        asset:
            Asset name.

        version:
            Version name.

        cache_dir:
            Path to the cache directory.

        overwrite:
            Whether to overwrite existing file in cache.

        url:
            URL to the gypsum compatible API.

    Returns:
        Dictionary containing the summary for this version, with the following fields:
        - ``upload_user_id``, string containing the identity of the uploader.
        - ``upload_start``, a POSIXct object containing the upload start time.
        - ``upload_finish``, a POSIXct object containing the upload finish time.
        - ``on_probation`` (optional), a logical scalar indicating whether the upload is probational.
            If missing, this can be assumed to be False.
    """
    _out = _fetch_cacheable_json(
        project,
        asset,
        version,
        "..summary",
        cache=cache_dir,
        overwrite=overwrite,
        url=url,
    )

    _out["upload_start"] = _cast_datetime(_out["upload_start"])
    _out["upload_finish"] = _cast_datetime(_out["upload_finish"])

    if "on_probation" in _out:
        if _out["on_probation"] is True and cache_dir is not None:
            _out_path = os.path.join(
                cache_dir, BUCKET_CACHE_NAME, project, asset, version, "..summary"
            )
            os.unlink(_out_path)

    return _out


def fetch_usage(project: str, url: str = rest_url()) -> int:
    """Fetch the quota usage for a project.

    See Also:
        :py:func:`~gypsum_client.refresh_operations.refresh_usage`,
        to refresh usage details.

    Example:

        .. code-block:: python

            usage = fetch_usage("test-R")

    Args:
        project:
            Project name.

        url:
            URL to the gypsum compatible API.

    Returns:
        Quota usage for the project, in bytes.
    """
    _usage = _fetch_json(f"{project}/..usage", url=url)
    return _usage["total"]



import os
import time
from typing import Optional, Union

import requests
from filelock import FileLock

from ._github import github_access_token
from ._utils import _is_interactive, _remove_slash_url
from .cache_directory import cache_directory
from .config import REQUESTS_MOD
from .rest_url import rest_url

__author__ = "Jayaram Kancherla"
__copyright__ = "Jayaram Kancherla"
__license__ = "MIT"

TOKEN_CACHE = {}


def _token_cache_path(cache_dir):
    return os.path.join(cache_dir, "credentials", "token.txt")


def access_token(
    full: bool = False,
    request: bool = True,
    cache_dir: Optional[str] = cache_directory(),
    token_expiration_limit: int = 10,
) -> Optional[Union[str, dict]]:
    """Get GitHub access token for authentication to the gypsum API's.

    Example:

        .. code-block:: python

            token = access_token()

    Args:
        full:
            Whether to return the full token details.
            Defaults to False, only ``token`` is returned.

        request:
            Whether to request a new token if no token is found or the
            current token is expired. Defaults to True.

        cache_dir:
            Path to the cache directory to store tokens.
            Can be set to `None`, indicating token is not cached to disk.

        token_expiration_limit:
            Integer specifying the number of seconds until the token expires.

    Returns:
        If `full=False` A string specifying the GitHub token to
        access gypsum's resources.

        If `full=True` retuns a dicionary containing the full token details.
    """
    global TOKEN_CACHE

    def _token_func(x):
        return x["token"] if not full else x

    in_memory = TOKEN_CACHE.get("auth_info", None)
    if in_memory is not None:
        if in_memory["expires"] > time.time() + token_expiration_limit:
            return _token_func(in_memory)
        else:
            TOKEN_CACHE["auth_info"] = None

    if cache_dir is not None:
        cache_path = _token_cache_path(cache_dir)

        if os.path.exists(cache_path):
            _lock = FileLock(cache_path + ".LOCK")
            with _lock:
                with open(cache_path, "r") as file:
                    dump = file.read().splitlines()

            if len(dump) > 0:
                exp = float(dump[2])
                if exp > time.time() + token_expiration_limit:
                    info = {"token": dump[0], "name": dump[1], "expires": exp}
                    TOKEN_CACHE["auth_info"] = info
                    return _token_func(info)
            else:
                os.unlink(cache_path)

    if request:
        payload = set_access_token(cache_dir=cache_dir)
        return _token_func(payload) if payload else None
    else:
        return None


TOKEN_AUTO = "auto"


def set_access_token(
    token: str = TOKEN_AUTO,
    app_url: str = rest_url(),
    app_key: Optional[str] = None,
    app_secret: Optional[str] = None,
    github_url: str = "https://api.github.com",
    user_agent: Optional[str] = None,
    cache_dir: Optional[str] = cache_directory(),
) -> dict:
    """Set GitHub access token for authentication to the gypsum API's.

    Args:
        token:
            A String containing Github's personal access token.

        app_url:
            URL to the gypsum REST API.

        app_key:
            Key to the GitHub oauth app.

        app_secret:
            Secret to the GitHub oauth app.

        github_url:
            URL to GitHub's API.

        user_agent:
            Specify the user agent for queries to various endpoints.

        cache_dir:
            Path to the cache directory to store tokens.
            Defaults to None, indicating token is not cached to disk.

    Returns:
        Dictionary containing the following keys:
        - ``token``, a string containing the token.
        - ``name``, the name of the GitHub user authenticated by the token.
        - ``expires``, the Unix time at which the token expires.
    """
    global TOKEN_CACHE

    cache_path = None
    if cache_dir is not None:
        cache_path = _token_cache_path(cache_dir)

    if token is None:
        TOKEN_CACHE["auth_info"] = None
        if cache_path is not None:
            os.unlink(cache_path)

        return

    if token is TOKEN_AUTO:
        if not _is_interactive():
            raise Exception(
                "Running in non-interactive mode. Set the token manually using `set_access_token`."
            )

        if not app_key or not app_secret:
            _url = f"{_remove_slash_url(app_url)}/credentials/github-app"
            headers = {}
            if user_agent:
                headers["User-Agent"] = user_agent

            r = requests.get(_url, headers=headers, verify=REQUESTS_MOD["verify"])
            try:
                r.raise_for_status()
            except Exception as e:
                raise Exception(
                    f"Failed to get access credentials from gypsum, {r.status_code} and reason: {r.text}"
                ) from e

            _info = r.json()
            app_key = _info["id"]
            app_secret = _info["secret"]

        token = github_access_token(
            client_id=app_key,
            client_secret=app_secret,
            authorization_url="https://github.com/login/oauth/authorize",
            token_url="https://github.com/login/oauth/access_token",
        )

    if token is None:
        raise ValueError("'token' cannot be 'None'.")

    headers = {}
    if user_agent:
        headers["User-Agent"] = user_agent

    headers["Authorization"] = f"Bearer {token}"

    token_req = requests.get(
        f"{_remove_slash_url(github_url)}/user",
        headers=headers,
        verify=REQUESTS_MOD["verify"],
    )
    try:
        token_req.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Failed to access token from GitHub, {token_req.status_code} and reason: {token_req.text}"
        ) from e

    token_resp = token_req.json()
    name = token_resp["login"]
    expires_header = token_req.headers.get("github-authentication-token-expiration")

    expiry = float("inf")
    if expires_header is not None:
        expiry = float(expires_header.split(" ")[0])

    if cache_dir is not None:
        cache_path = _token_cache_path(cache_dir)
        os.makedirs(os.path.dirname(cache_path), exist_ok=True)

        _lock = FileLock(cache_path + ".LOCK")
        with _lock:
            with open(cache_path, "w") as file:
                file.write("\n".join([token, name, str(expiry)]))

        os.chmod(
            cache_path, 0o600
        )  # prevent anyone else from reading this on shared file systems.

    vals = {"token": token, "name": name, "expires": expiry}
    TOKEN_CACHE["auth_info"] = vals
    return vals
